import math
import os
from pathlib import Path
import glob

import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.legend_handler import HandlerTuple
import pandas as pd
from astropy import constants as const

import artistools as at
from .lightcurve import *


define_colours_list = ['k', 'tab:blue', 'tab:red', 'tab:green', 'purple', 'tab:orange', 'tab:pink', 'tab:gray', 'gold',
                       'tab:cyan', 'darkblue', 'darkgreen', 'maroon', 'mediumvioletred', 'saddlebrown', 'darkslategrey',
                       'bisque', 'yellow', 'k', 'tab:blue', 'tab:red', 'tab:green', 'purple',
                       'tab:orange', 'tab:pink', 'tab:gray', 'gold', 'tab:cyan', 'darkblue', 'bisque', 'yellow', 'k',
                       'tab:blue', 'tab:red', 'tab:green', 'purple', 'tab:orange', 'tab:pink', 'tab:gray', 'gold',
                       'tab:cyan',
                       'darkblue', 'bisque', 'yellow', 'k', 'tab:blue', 'tab:red', 'tab:green', 'purple', 'tab:orange',
                       'tab:pink', 'tab:gray', 'gold', 'tab:cyan', 'darkblue', 'bisque', 'yellow', 'k', 'tab:blue',
                       'tab:red',
                       'tab:green', 'purple', 'tab:orange', 'tab:pink', 'tab:gray', 'gold', 'tab:cyan', 'darkblue',
                       'bisque',
                       'yellow', 'k', 'tab:blue', 'tab:red', 'tab:green', 'purple', 'tab:orange', 'tab:pink',
                       'tab:gray',
                       'gold', 'tab:cyan', 'darkblue', 'bisque', 'yellow', 'k', 'tab:blue', 'tab:red', 'tab:green',
                       'purple',
                       'tab:orange', 'tab:pink', 'tab:gray', 'gold', 'tab:cyan', 'darkblue', 'bisque', 'yellow', 'k',
                       'tab:blue', 'tab:red', 'tab:green', 'purple', 'tab:orange', 'tab:pink', 'tab:gray', 'gold',
                       'tab:cyan',
                       'darkblue', 'bisque', 'yellow']

define_colours_list2 = ['gray', 'lightblue', 'pink', 'yellowgreen', 'mediumorchid', 'sandybrown', 'plum', 'lightgray',
                        'wheat', 'paleturquoise', 'royalblue', 'springgreen', 'r', 'deeppink', 'sandybrown', 'teal']


def get_angle_stuff(modelpath, args):
    modelpath = Path(modelpath)
    viewing_angles = None
    viewing_angle_data = False
    if len(glob.glob(str(modelpath / '*_res.out'))) > 1:
        viewing_angle_data = True

    if args.plotvspecpol and os.path.isfile(modelpath / 'vpkt.txt'):
        angles = args.plotvspecpol
    elif args.plotviewingangle and args.plotviewingangle[0] == -1 and viewing_angle_data:
        angles = np.arange(0, 100, 1, dtype=int)
    elif args.plotviewingangle and viewing_angle_data:
        angles = args.plotviewingangle
    elif args.calculate_costheta_phi_from_viewing_angle_numbers and \
            args.calculate_costheta_phi_from_viewing_angle_numbers[0] == -1:
        viewing_angles = np.arange(0, 100, 1, dtype=int)
        calculate_costheta_phi_for_viewing_angles(viewing_angles, modelpath)
    elif args.calculate_costheta_phi_from_viewing_angle_numbers:
        viewing_angles = args.calculate_costheta_phi_from_viewing_angle_numbers
        calculate_costheta_phi_for_viewing_angles(viewing_angles, modelpath)
    else:
        angles = [None]

    angle_definition = None
    if angles[0] is not None and not args.plotvspecpol:
        angle_definition = calculate_costheta_phi_for_viewing_angles(angles, modelpath)
        if args.average_every_tenth_viewing_angle:
            for key in angle_definition.keys():
                costheta_label = angle_definition[key].split(',')[0]
                angle_definition[key] = costheta_label

    return angles, viewing_angles, angle_definition


def get_viewinganglebin_definitions():
    costheta_viewing_angle_bins = ['-1.0 \u2264 cos(\u03B8) < -0.8', '-0.8 \u2264 cos(\u03B8) < -0.6',
                                   '-0.6 \u2264 cos(\u03B8) < -0.4', '-0.4 \u2264 cos(\u03B8) < -0.2',
                                   '-0.2 \u2264 cos(\u03B8) < 0', '0 \u2264 cos(\u03B8) < 0.2',
                                   '0.2 \u2264 cos(\u03B8) < 0.4', '0.4 \u2264 cos(\u03B8) < 0.6',
                                   '0.6 \u2264 cos(\u03B8) < 0.8', '0.8 \u2264 cos(\u03B8) < 1']
    phi_viewing_angle_bins = ['0 \u2264 \u03D5 < \u03c0/5', '\u03c0/5 \u2264 \u03D5 < 2\u03c0/5',
                              '2\u03c0/5 \u2264 \u03D5 < 3\u03c0/5', '3\u03c0/5 \u2264 \u03D5 < 4\u03c0/5',
                              '4\u03c0/5 \u2264 \u03D5 < \u03c0', '9\u03c0/5 < \u03D5 < 2\u03c0',
                              '8\u03c0/5 < \u03D5 \u2264 9\u03c0/5', '7\u03c0/5 < \u03D5 \u2264 8\u03c0/5',
                              '6\u03c0/5 < \u03D5 \u2264 7\u03c0/5', '\u03c0 < \u03D5 \u2264 6\u03c0/5']

    # label orders changed so that bins are in order. Not used yet.
    # phi_viewing_angle_bins_reordered = ['0 \u2264 \u03D5 < \u03c0/5', '\u03c0/5 \u2264 \u03D5 < 2\u03c0/5',
    #                                     '2\u03c0/5 \u2264 \u03D5 < 3\u03c0/5', '3\u03c0/5 \u2264 \u03D5 < 4\u03c0/5',
    #                                     '4\u03c0/5 \u2264 \u03D5 < \u03c0', '\u03c0 < \u03D5 \u2264 6\u03c0/5',
    #                                     '6\u03c0/5 < \u03D5 \u2264 7\u03c0/5', '7\u03c0/5 < \u03D5 \u2264 8\u03c0/5',
    #                                     '8\u03c0/5 < \u03D5 \u2264 9\u03c0/5', '9\u03c0/5 < \u03D5 < 2\u03c0']
    return costheta_viewing_angle_bins, phi_viewing_angle_bins


def calculate_costheta_phi_for_viewing_angles(viewing_angles, modelpath):
    modelpath = Path(modelpath)
    if os.path.isfile(modelpath / 'absorptionpol_res_99.out') \
            and os.path.isfile(modelpath / 'absorptionpol_res_100.out'):
        print("Too many viewing angle bins (MABINS) for this method to work, it only works for MABINS = 100")
        exit()
    elif os.path.isfile(modelpath / 'light_curve_res.out'):
        angle_definition = {}

        costheta_viewing_angle_bins, phi_viewing_angle_bins = get_viewinganglebin_definitions()

        for angle in viewing_angles:
            MABINS = 100
            phibins = int(math.sqrt(MABINS))
            costheta_index = angle // phibins
            phi_index = angle % phibins

            angle_definition[angle] = f'{costheta_viewing_angle_bins[costheta_index]}, {phi_viewing_angle_bins[phi_index]}'
            print(f"{angle:4d}   {costheta_viewing_angle_bins[costheta_index]}   {phi_viewing_angle_bins[phi_index]}")

        return angle_definition
    else:
        print("Too few viewing angle bins (MABINS) for this method to work, it only works for MABINS = 100")
        exit()


def save_viewing_angle_data_for_plotting(band_name, modelname, args):
    if args.save_viewing_angle_peakmag_risetime_delta_m15_to_file:
        np.savetxt(band_name + "band_" + f'{modelname}' + "_viewing_angle_data.txt",
                   np.c_[args.band_peakmag_polyfit, args.band_risetime_polyfit, args.band_deltam15_polyfit],
                   delimiter=' ', header='peak_mag_polyfit risetime_polyfit deltam15_polyfit', comments='')

    elif (args.save_angle_averaged_peakmag_risetime_delta_m15_to_file
          or args.make_viewing_angle_peakmag_risetime_scatter_plot
          or args.make_viewing_angle_peakmag_delta_m15_scatter_plot):

        args.band_risetime_angle_averaged_polyfit.append(args.band_risetime_polyfit)
        args.band_peakmag_angle_averaged_polyfit.append(args.band_peakmag_polyfit)
        args.band_delta_m15_angle_averaged_polyfit.append(args.band_deltam15_polyfit)

    args.band_risetime_polyfit = []
    args.band_peakmag_polyfit = []
    args.band_deltam15_polyfit = []

    # if args.magnitude and not (
    #         args.calculate_peakmag_risetime_delta_m15 or args.save_angle_averaged_peakmag_risetime_delta_m15_to_file
    #         or args.save_viewing_angle_peakmag_risetime_delta_m15_to_file or args.test_viewing_angle_fit
    #         or args.make_viewing_angle_peakmag_risetime_scatter_plot or
    #         args.make_viewing_angle_peakmag_delta_m15_scatter_plot or args.plotviewingangle):
    #     plt.plot(time, magnitude, label=modelname, color=colours[modelnumber], linewidth=3)


def write_viewing_angle_data(band_name, modelnames, args):
    if (args.save_angle_averaged_peakmag_risetime_delta_m15_to_file
            or args.make_viewing_angle_peakmag_risetime_scatter_plot
            or args.make_viewing_angle_peakmag_delta_m15_scatter_plot):
        np.savetxt(band_name + "band_" + f'{modelnames[0]}' + "_angle_averaged_all_models_data.txt",
                   np.c_[modelnames, args.band_risetime_angle_averaged_polyfit, args.band_peakmag_angle_averaged_polyfit,
                         args.band_delta_m15_angle_averaged_polyfit],
                   delimiter=' ', fmt='%s',
                   header="object " + str(band_name) + "_band_risetime " + str(band_name) + "_band_peakmag " + str(
                       band_name) + "_band_deltam15 ", comments='')


def calculate_peak_time_mag_deltam15(time, magnitude, modelname, angle, key, args, filternames_conversion_dict=None):
    """Calculating band peak time, peak magnitude and delta m15"""
    if args.timemin is None or args.timemax is None:
        print("Trying to calculate peak time / dm15 / rise time with no time range. "
              "This will give a stupid result. Specify args.timemin and args.timemax")
        quit()

    fxfit, xfit = lightcurve_polyfit(time, magnitude, args)

    def match_closest_time_polyfit(reftime_polyfit):
        return str("{}".format(min([float(x) for x in xfit], key=lambda x: abs(x - reftime_polyfit))))

    index_min = np.argmin(fxfit)
    tmax_polyfit = xfit[index_min]
    time_after15days_polyfit = match_closest_time_polyfit(tmax_polyfit + 15)
    for ii, xfits in enumerate(xfit):
        if float(xfits) == float(time_after15days_polyfit):
            index_after_15_days = ii

    mag_after15days_polyfit = fxfit[index_after_15_days]
    print(f'{key}_max polyfit = {min(fxfit)} at time = {tmax_polyfit}')
    print(f'deltam15 polyfit = {min(fxfit) - mag_after15days_polyfit}')

    args.band_risetime_polyfit.append(tmax_polyfit)
    args.band_peakmag_polyfit.append(min(fxfit))
    args.band_deltam15_polyfit.append((min(fxfit) - mag_after15days_polyfit) * -1)

    # Plotting the lightcurves for all viewing angles specified in the command line along with the
    # polynomial fit and peak mag, risetime to peak and delta m15 marked on the plots to check the
    # fit is working correctly
    if args.test_viewing_angle_fit:
        make_plot_test_viewing_angle_fit(time, magnitude, xfit, fxfit, filternames_conversion_dict, key,
                                         mag_after15days_polyfit, tmax_polyfit, time_after15days_polyfit,
                                         modelname, angle)


def lightcurve_polyfit(time, magnitude, args):
    zfit = np.polyfit(x=time, y=magnitude, deg=10)
    xfit = np.linspace(args.timemin + 0.5, args.timemax - 0.5, num=1000)

    # Taking line_min and line_max from the limits set for the lightcurve being plotted
    fxfit = []
    for j, _ in enumerate(xfit):
        fxfit.append(zfit[0] * (xfit[j] ** 10) + zfit[1] * (xfit[j] ** 9) + zfit[2] * (xfit[j] ** 8) +
                     zfit[3] * (xfit[j] ** 7) + zfit[4] * (xfit[j] ** 6) + zfit[5] * (xfit[j] ** 5) +
                     zfit[6] * (xfit[j] ** 4) + zfit[7] * (xfit[j] ** 3) + zfit[8] * (xfit[j] ** 2) +
                     zfit[9] * (xfit[j]) + zfit[10])
        # polynomial with 10 degrees of freedom used here but change as required if it improves the fit
    return fxfit, xfit


def make_plot_test_viewing_angle_fit(time, magnitude, xfit, fxfit, filternames_conversion_dict, key,
                                     mag_after15days_polyfit, tmax_polyfit, time_after15days_polyfit,
                                     modelname, angle):
    plt.plot(time, magnitude)
    plt.plot(xfit, fxfit)

    if key in filternames_conversion_dict:
        plt.ylabel(f'{filternames_conversion_dict[key]} Magnitude')
    else:
        plt.ylabel(f'{key} Magnitude')

    plt.xlabel('Time Since Explosion [d]')
    plt.gca().invert_yaxis()
    plt.xlim(0, 40)
    plt.minorticks_on()
    plt.tick_params(axis='both', which='minor', top=True, right=True, length=5, width=2, labelsize=12)
    plt.tick_params(axis='both', which='major', top=True, right=True, length=8, width=2, labelsize=12)
    plt.axhline(y=min(fxfit), color="black", linestyle="--")
    plt.axhline(y=mag_after15days_polyfit, color="black", linestyle="--")
    plt.axvline(x=tmax_polyfit, color="black", linestyle="--")
    plt.axvline(x=float(time_after15days_polyfit), color="black", linestyle="--")
    print("time after 15 days polyfit = ", time_after15days_polyfit)
    plt.tight_layout()
    plt.savefig(f'{key}' + "_band_" + f'{modelname}' + "_viewing_angle" + str(angle) + ".png")
    plt.close()


def set_scatterplot_plotkwargs(modelnumber, args):
    plotkwargsviewingangles = {}
    plotkwargsviewingangles['marker'] = 'x'
    plotkwargsviewingangles['zorder'] = 0
    plotkwargsviewingangles['alpha'] = 0.8
    if args.colorbarcostheta or args.colorbarphi:
        update_plotkwargs_for_viewingangle_colorbar(plotkwargsviewingangles, args)
    else:
        if args.color:
            plotkwargsviewingangles['color'] = args.color[modelnumber]
        else:
            plotkwargsviewingangles['color'] = define_colours_list2[modelnumber]

    plotkwargsangleaveraged = {}
    plotkwargsangleaveraged['marker'] = 'o'
    plotkwargsangleaveraged['zorder'] = 10
    plotkwargsangleaveraged['edgecolor'] = 'k'
    plotkwargsangleaveraged['s'] = 120
    if args.color:
        plotkwargsangleaveraged['color'] = args.color[modelnumber]
    else:
        plotkwargsangleaveraged['color'] = define_colours_list[modelnumber]

    if args.colorbarcostheta or args.colorbarphi:
        update_plotkwargs_for_viewingangle_colorbar(plotkwargsviewingangles, args)

    return plotkwargsviewingangles, plotkwargsangleaveraged


def update_plotkwargs_for_viewingangle_colorbar(plotkwargsviewingangles, args):

    costheta_viewing_angle_bins, phi_viewing_angle_bins = get_viewinganglebin_definitions()
    scaledmap = at.lightcurve.plotlightcurve.make_colorbar_viewingangles_colormap()

    angles = np.arange(0, 100)
    angle_definition = calculate_costheta_phi_for_viewing_angles(angles, args.modelpath[0])
    colors = []
    for angle in angles:
        _, colorindex = at.lightcurve.plotlightcurve.get_viewinganglecolor_for_colorbar(
            angle_definition, angle, costheta_viewing_angle_bins, phi_viewing_angle_bins,
            scaledmap, plotkwargsviewingangles, args)
        colors.append(scaledmap.to_rgba(colorindex))
    plotkwargsviewingangles['color'] = colors
    return plotkwargsviewingangles


def set_scatterplot_plot_params(args):
    if not args.colouratpeak:
        plt.gca().invert_yaxis()
    plt.xlim(args.xmin, args.xmax)
    plt.ylim(args.ymin, args.ymax)
    plt.minorticks_on()
    plt.tick_params(axis='both', which='minor', top=False, right=False, length=5, width=2, labelsize=12)
    plt.tick_params(axis='both', which='major', top=False, right=False, length=8, width=2, labelsize=12)
    plt.tight_layout()

    if args.colorbarcostheta or args.colorbarphi:
        costheta_viewing_angle_bins, phi_viewing_angle_bins = get_viewinganglebin_definitions()
        scaledmap = at.lightcurve.plotlightcurve.make_colorbar_viewingangles_colormap()
        at.lightcurve.plotlightcurve.make_colorbar_viewingangles(phi_viewing_angle_bins, scaledmap, args)

# COMBINED WITH DM15 plotting function now ###
# def make_viewing_angle_peakmag_risetime_scatter_plot(modelnames, key, args):
#     for ii, modelname in enumerate(modelnames):
#         viewing_angle_plot_data = pd.read_csv(key + "band_" + f'{modelname}' + "_viewing_angle_data.txt",
#                                               delimiter=" ")
#         band_peak_mag_viewing_angles = viewing_angle_plot_data["peak_mag_polyfit"].values
#         band_risetime_viewing_angles = viewing_angle_plot_data["risetime_polyfit"].values
#
#         plotkwargsviewingangles, plotkwargsangleaveraged = set_scatterplot_plotkwargs(ii, args)
#
#         a0 = plt.scatter(band_risetime_viewing_angles, band_peak_mag_viewing_angles, **plotkwargsviewingangles)
#         if not args.noangleaveraged:
#             p0 = plt.scatter(args.band_risetime_angle_averaged_polyfit[ii], args.band_peakmag_angle_averaged_polyfit[ii],
#                              **plotkwargsangleaveraged)
#             args.plotvalues.append((a0, p0))
#         else:
#             args.plotvalues.append((a0, a0))
#         if not args.noerrorbars:
#             plt.errorbar(args.band_risetime_angle_averaged_polyfit[ii], args.band_peakmag_angle_averaged_polyfit[ii],
#                          xerr=np.std(band_risetime_viewing_angles),
#                          yerr=np.std(band_peak_mag_viewing_angles), ecolor=define_colours_list[ii], capsize=2)
#
#     if not args.nolegend:
#         plt.legend(args.plotvalues, modelnames, numpoints=1, handler_map={tuple: HandlerTuple(ndivide=None)},
#                    loc='upper left', fontsize=8, ncol=2, columnspacing=1, frameon=False)
#     plt.xlabel('Rise Time in Days', fontsize=14)
#     if args.filter:
#         ylabel = 'Peak ' + key + ' Band Magnitude'
#     else:
#         ylabel = 'Peak Magnitude'
#     plt.ylabel(ylabel, fontsize=14)
#     if args.title:
#         plt.title(f"{at.get_model_name(args.modelpath[0])}")
#     set_scatterplot_plot_params(args)
#     if args.show:
#         plt.show()
#     plt.savefig(key + "_band_" + f'{modelnames[0]}' + "_viewing_angle_peakmag_risetime_scatter_plot.pdf", format="pdf")
#     print("saving " + key + "_band_" + f'{modelnames[0]}' + "_viewing_angle_peakmag_risetime_scatter_plot.pdf")
#     plt.close()


def make_viewing_angle_risetime_peakmag_delta_m15_scatter_plot(modelnames, key, args):
    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True,
                           figsize=(8, 6), tight_layout={"pad": 0.5, "w_pad": 1.5, "h_pad": 0.3})

    for ii, modelname in enumerate(modelnames):
        viewing_angle_plot_data = pd.read_csv(key + "band_" + f'{modelname}' + "_viewing_angle_data.txt",
                                              delimiter=" ")

        band_peak_mag_viewing_angles = viewing_angle_plot_data["peak_mag_polyfit"].values
        band_delta_m15_viewing_angles = viewing_angle_plot_data["deltam15_polyfit"].values
        band_risetime_viewing_angles = viewing_angle_plot_data["risetime_polyfit"].values

        plotkwargsviewingangles, plotkwargsangleaveraged = set_scatterplot_plotkwargs(ii, args)

        if args.make_viewing_angle_peakmag_delta_m15_scatter_plot:
            xvalues_viewingangles = band_delta_m15_viewing_angles
        if args.make_viewing_angle_peakmag_risetime_scatter_plot:
            xvalues_viewingangles = band_risetime_viewing_angles

        a0 = ax.scatter(xvalues_viewingangles, band_peak_mag_viewing_angles, **plotkwargsviewingangles)

        if not args.noangleaveraged:
            if args.make_viewing_angle_peakmag_delta_m15_scatter_plot:
                xvalues_angleaveraged = args.band_delta_m15_angle_averaged_polyfit[ii]
            if args.make_viewing_angle_peakmag_risetime_scatter_plot:
                xvalues_angleaveraged = args.band_risetime_angle_averaged_polyfit[ii]

            p0 = ax.scatter(xvalues_angleaveraged, args.band_peakmag_angle_averaged_polyfit[ii],
                             **plotkwargsangleaveraged)
            args.plotvalues.append((a0, p0))
        else:
            args.plotvalues.append((a0, a0))
        if not args.noerrorbars:
            if args.color:
                ecolor = args.color
            else:
                ecolor = define_colours_list

            ax.errorbar(xvalues_angleaveraged, args.band_peakmag_angle_averaged_polyfit[ii],
                         xerr=np.std(xvalues_viewingangles),
                         yerr=np.std(band_peak_mag_viewing_angles), ecolor=ecolor[ii], capsize=2)

    if args.label:
        linelabels = args.label
    else:
        linelabels = modelnames

    # a0, datalabel = at.lightcurve.get_sn_sample_bol()
    # a0, datalabel = at.lightcurve.plot_phillips_relation_data()
    # args.plotvalues.append((a0, a0))
    # linelabels.append(datalabel)

    if not args.nolegend:
        ax.legend(args.plotvalues, linelabels, numpoints=1, handler_map={tuple: HandlerTuple(ndivide=None)},
                   loc='upper right', fontsize='x-small', ncol=args.ncolslegend, columnspacing=1, frameon=False)
    # ax.set_xlabel(r'Decline Rate ($\Delta$m$_{15}$)', fontsize=14)

    if args.make_viewing_angle_peakmag_delta_m15_scatter_plot:
        xlabel = r'$\Delta$m$_{15}$' + f'({key})'
    if args.make_viewing_angle_peakmag_risetime_scatter_plot:
        xlabel = 'Rise Time [days]'

    ax.set_xlabel(xlabel, fontsize=14)
    # ax.set_ylabel('Peak ' + key + ' Band Magnitude', fontsize=14)
    ax.set_ylabel(rf'M$_{{\mathrm{{{key}}}}}$, max', fontsize=14)
    set_scatterplot_plot_params(args)

    if args.make_viewing_angle_peakmag_delta_m15_scatter_plot:
        filename = fr'{key}_band_{modelnames[0]}_dm15_peakmag.pdf'
    if args.make_viewing_angle_peakmag_risetime_scatter_plot:
        filename = fr'{key}_band_{modelnames[0]}_risetime_peakmag.pdf'
    plt.savefig(filename, format="pdf")
    print(f"saving {filename}")
    plt.close()


def make_peak_colour_viewing_angle_plot(args):
    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=True,
                           figsize=(8, 6), tight_layout={"pad": 0.5, "w_pad": 1.5, "h_pad": 0.3})

    for modelnumber, modelpath in enumerate(args.modelpath):
        modelname = at.get_model_name(modelpath)

        bands = [args.filter[0], args.filter[1]]

        data = {}

        datafilename = bands[0] + "band_" + f'{modelname}' + "_viewing_angle_data.txt"
        viewing_angle_plot_data = pd.read_csv(datafilename, delimiter=" ")
        data[f"{bands[0]}max"] = viewing_angle_plot_data["peak_mag_polyfit"].values
        data[f"time_{bands[0]}max"] = viewing_angle_plot_data["risetime_polyfit"].values

        # Get brightness in second band at time of peak in first band
        if len(data[f"time_{bands[0]}max"]) != 100:
            print(f"All 100 angles are not in file {datafilename}. Quitting")
            quit()

        second_band_brightness = second_band_brightness_at_peak_first_band(data, bands, modelpath, modelnumber, args)

        data[f"{bands[1]}at{bands[0]}max"] = second_band_brightness

        data = pd.DataFrame(data)
        data['peakcolour'] = data[f"{bands[0]}max"] - data[f"{bands[1]}at{bands[0]}max"]
        print(data['peakcolour'], data[f"{bands[0]}max"], data[f"{bands[1]}at{bands[0]}max"])

        plotkwargsviewingangles, _ = set_scatterplot_plotkwargs(modelnumber, args)
        plotkwargsviewingangles['label'] = modelname
        ax.scatter(data['peakcolour'], data[f"{bands[0]}max"], **plotkwargsviewingangles)

    sn_data, label = get_phillips_relation_data()
    ax.errorbar(x=sn_data['(B-V)Bmax'], y=sn_data['MB'], xerr=sn_data['err_(B-V)Bmax'], yerr=sn_data['err_MB'],
                color='k', alpha=0.9, marker='.', capsize=2, label=label, ls='None', zorder=-1)

    ax.legend(loc='upper right', fontsize=8, ncol=1, columnspacing=1, frameon=False)
    ax.set_xlabel(f'{bands[0]}-{bands[1]} at {bands[0]}max', fontsize=14)
    ax.set_ylabel(f'{bands[0]}max', fontsize=14)
    set_scatterplot_plot_params(args)
    plotname = f'plotviewinganglecolour{bands[0]}-{bands[1]}.pdf'
    plt.savefig(plotname, format="pdf")
    print(f"saving {plotname}")
    plt.close()


@at.diskcache(savezipped=True)
def second_band_brightness_at_peak_first_band(data, bands, modelpath, modelnumber, args):
    second_band_brightness = []
    for anglenumber, time in enumerate(data[f"time_{bands[0]}max"]):
        lightcurve_data = generate_band_lightcurve_data(modelpath, args, anglenumber, modelnumber=modelnumber)
        time, brightness_in_mag = get_band_lightcurve(lightcurve_data, bands[1], args)

        fxfit, xfit = lightcurve_polyfit(time, brightness_in_mag, args)

        closest_list_time_to_first_band_peak \
            = at.match_closest_time(reftime=data[f"time_{bands[0]}max"][anglenumber], searchtimes=xfit)

        for ii, xfits in enumerate(xfit):
            if float(xfits) == float(closest_list_time_to_first_band_peak):
                index_at_max = ii
                break

        brightness_in_second_band_at_first_band_peak = fxfit[index_at_max]
        print(brightness_in_second_band_at_first_band_peak)
        second_band_brightness.append(brightness_in_second_band_at_first_band_peak)

    return second_band_brightness


def peakmag_risetime_declinerate_init(modelpaths, filternames_conversion_dict, args):

    # if args.calculate_peak_time_mag_deltam15_bool:  # If there's viewing angle scatter plot stuff define some arrays
    args.plotvalues = []  # a0 and p0 values for viewing angle scatter plots

    args.band_risetime_polyfit = []
    args.band_peakmag_polyfit = []
    args.band_deltam15_polyfit = []

    args.band_risetime_angle_averaged_polyfit = []
    args.band_peakmag_angle_averaged_polyfit = []
    args.band_delta_m15_angle_averaged_polyfit = []

    modelnames = []  # save names of models

    for modelnumber, modelpath in enumerate(modelpaths):
        modelpath = Path(modelpath)

        if not args.filter:
            if args.plotviewingangle:
                lcname = 'light_curve_res.out'
            else:
                lcname = 'light_curve.out'
            lcpath = at.firstexisting([lcname + '.xz', lcname + '.gz', lcname], path=modelpath)
            print(f"Reading {lcname}")
            lightcurve_data = at.lightcurve.readfile(lcpath, args)

        # check if doing viewing angle stuff, and if so define which data to use
        angles, viewing_angles, angle_definition = get_angle_stuff(modelpath, args)
        if not args.filter and args.plotviewingangle:
            lcdataframes = lightcurve_data

        for index, angle in enumerate(angles):

            modelname = at.get_model_name(modelpath)
            modelnames.append(modelname)  # save for later
            print(f'Reading spectra: {modelname}')
            if args.filter:
                lightcurve_data = generate_band_lightcurve_data(modelpath, args, angle, modelnumber=modelnumber)
                plottinglist = args.filter
            elif args.plotviewingangle:
                lightcurve_data = lcdataframes[angle]
            if not args.filter:
                plottinglist = ['lightcurve']

            for plotnumber, band_name in enumerate(plottinglist):
                if args.filter:
                    time, brightness = get_band_lightcurve(lightcurve_data, band_name, args)
                else:
                    lightcurve_data = lightcurve_data.loc[(lightcurve_data['time'] > args.timemin) &
                                                          (lightcurve_data['time'] < args.timemax)]

                    lightcurve_data['mag'] = 4.74 - (2.5 * np.log10((lightcurve_data['lum'] * 3.826e33)
                                                                    / const.L_sun.to('erg/s').value))

                    lightcurve_data = lightcurve_data.replace([np.inf, -np.inf], 0)
                    brightness = [mag for mag in lightcurve_data['mag'] if mag != 0]  # drop times with 0 brightness
                    time = [t for t, mag in zip(lightcurve_data['time'], lightcurve_data['mag']) if mag != 0]

                # Calculating band peak time, peak magnitude and delta m15
                if args.calculate_peak_time_mag_deltam15_bool:
                    calculate_peak_time_mag_deltam15(time, brightness, modelname, angle, band_name,
                                                     args, filternames_conversion_dict=filternames_conversion_dict)

        # Saving viewing angle data so it can be read in and plotted later on without re-running the script
        #    as it is quite time consuming
        if args.calculate_peak_time_mag_deltam15_bool:
            save_viewing_angle_data_for_plotting(plottinglist[0], modelname, args)

    # Saving all this viewing angle info for each model to a file so that it is available to plot if required again
    # as it takes relatively long to run this for all viewing angles
    if args.calculate_peak_time_mag_deltam15_bool:
        write_viewing_angle_data(plottinglist[0], modelnames, args)

    # if args.make_viewing_angle_peakmag_risetime_scatter_plot:
    #     make_viewing_angle_peakmag_risetime_scatter_plot(modelnames, plottinglist[0], args)
    #     return

    if args.make_viewing_angle_peakmag_delta_m15_scatter_plot or args.make_viewing_angle_peakmag_risetime_scatter_plot:
        make_viewing_angle_risetime_peakmag_delta_m15_scatter_plot(modelnames, plottinglist[0], args)
        return


def plot_viewanglebrightness_at_fixed_time(modelpath, args):
    fig, axis = plt.subplots(
        nrows=1, ncols=1, sharey=True, figsize=(8, 5), tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    angles, viewing_angles, angle_definition = at.lightcurve.get_angle_stuff(modelpath, args)

    costheta_viewing_angle_bins, phi_viewing_angle_bins = at.lightcurve.get_viewinganglebin_definitions()
    scaledmap = at.lightcurve.plotlightcurve.make_colorbar_viewingangles_colormap()

    plotkwargs = {}

    lcdataframes = at.lightcurve.readfile(modelpath / 'light_curve_res.out', args)

    timetoplot = at.match_closest_time(reftime=args.timedays, searchtimes=lcdataframes[0]['time'])
    print(timetoplot)

    for angleindex, lcdata in enumerate(lcdataframes):
        angle = angleindex
        plotkwargs, _ = at.lightcurve.plotlightcurve.get_viewinganglecolor_for_colorbar(
            angle_definition, angle, costheta_viewing_angle_bins, phi_viewing_angle_bins,
            scaledmap, plotkwargs, args)

        rowattime = lcdata.loc[lcdata['time'] == float(timetoplot)]
        brightness = (rowattime['lum'].item()) * 3.826e33
        if args.colorbarphi:
            xvalues = int(angleindex/10)
            xlabels = costheta_viewing_angle_bins
        if args.colorbarcostheta:
            xvalues = angleindex % 10
            xlabels = phi_viewing_angle_bins

        axis.scatter(xvalues, brightness, **plotkwargs)
        plt.xticks(ticks=np.arange(0, 10), labels=xlabels, rotation=30, ha='right')

    at.lightcurve.plotlightcurve.make_colorbar_viewingangles(phi_viewing_angle_bins, scaledmap, args)

    axis.set_xlabel('Angle bin')
    axis.set_ylabel('erg/s')
    axis.set_yscale('log')

    axis.set_title(f'time = {args.timedays} days')
    if args.show:
        plt.show()

    plotname = f"plotviewinganglebrightnessat{args.timedays}days.pdf"
    plt.savefig(plotname, format='pdf')
    print(f'Saved figure: {plotname}')



import errno
import gc
# from collections import namedtuple
import math
import os
import os.path
import time
from functools import lru_cache
from pathlib import Path

import artistools
import artistools.inputmodel.botyanski2017
import artistools.inputmodel.describeinputmodel
import artistools.inputmodel.makeartismodel
import artistools.inputmodel.rprocess_from_trajectory
import numpy as np
import pandas as pd


@lru_cache(maxsize=8)
def get_modeldata(inputpath=Path(), dimensions=None, get_abundances=False, derived_cols=False):
    """
    Read an artis model.txt file containing cell velocities, density, and abundances of radioactive nuclides.

    Arguments:
        - inputpath: either a path to model.txt file, or a folder containing model.txt
        - dimensions: number of dimensions in input file, or None for automatic
        - get_abundances: also read elemental abundances (abundances.txt) and
            merge with the output DataFrame

    Returns (dfmodel, t_model_init_days)
        - dfmodel: a pandas DataFrame with a row for each model grid cell
        - t_model_init_days: the time in days at which the snapshot is defined
    """

    assert dimensions in [1, 3, None]

    inputpath = Path(inputpath)

    if os.path.isdir(inputpath):
        modelpath = inputpath
        filename = artistools.firstexisting(['model.txt.xz', 'model.txt.gz', 'model.txt'], path=inputpath)
    elif os.path.isfile(inputpath):  # passed in a filename instead of the modelpath
        filename = inputpath
        modelpath = Path(inputpath).parent
    elif not inputpath.exists() and inputpath.parts[0] == 'codecomparison':
        modelpath = inputpath
        _, inputmodel, _ = modelpath.parts
        filename = Path(artistools.config['codecomparisonmodelartismodelpath'], inputmodel, 'model.txt')
    else:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), inputpath)

    headerrows = 0
    with artistools.zopen(filename, 'rt') as fmodel:
        gridcellcount = int(fmodel.readline())
        t_model_init_days = float(fmodel.readline())
        headerrows += 2
        t_model_init_seconds = t_model_init_days * 24 * 60 * 60

        filepos = fmodel.tell()
        # if the next line is a single float then the model is 3D
        try:
            vmax_cmps = float(fmodel.readline())  # velocity max in cm/s
            xmax_tmodel = vmax_cmps * t_model_init_seconds  # xmax = ymax = zmax
            headerrows += 1
            if dimensions is None:
                print("Detected 3D model file")
                dimensions = 3
            elif dimensions != 3:
                print(f" {dimensions} were specified but file appears to be 3D")
                assert False

        except ValueError:

            if dimensions is None:
                print("Detected 1D model file")
                dimensions = 1
            elif dimensions != 1:
                print(f" {dimensions} were specified but file appears to be 1D")
                assert False

            fmodel.seek(filepos)  # undo the readline() and go back

        columns = None
        filepos = fmodel.tell()
        line = fmodel.readline()
        if line.startswith('#'):
            headerrows += 1
            columns = line.lstrip('#').split()
        else:
            fmodel.seek(filepos)  # undo the readline() and go back

        ncols_file = len(fmodel.readline().split())
        if dimensions > 1:
            # columns split over two lines
            ncols_file += len(fmodel.readline().split())

        if columns is not None:
            assert ncols_file == len(columns)
        elif dimensions == 1:
            columns = ['inputcellid', 'velocity_outer', 'logrho', 'X_Fegroup', 'X_Ni56',
                       'X_Co56', 'X_Fe52', 'X_Cr48', 'X_Ni57', 'X_Co57'][:ncols_file]

        elif dimensions == 3:
            columns = ['inputcellid', 'inputpos_a', 'inputpos_b', 'inputpos_c', 'rho',
                       'X_Fegroup', 'X_Ni56', 'X_Co56', 'X_Fe52', 'X_Cr48', 'X_Ni57', 'X_Co57'][:ncols_file]

            # number of grid cell steps along an axis (same for xyz)
            ncoordgridx = int(round(gridcellcount ** (1. / 3.)))
            ncoordgridy = int(round(gridcellcount ** (1. / 3.)))
            ncoordgridz = int(round(gridcellcount ** (1. / 3.)))

            assert (ncoordgridx * ncoordgridy * ncoordgridz) == gridcellcount

    if dimensions == 1:
        dfmodel = pd.read_csv(
            filename, delim_whitespace=True, header=None, names=columns, skiprows=headerrows, nrows=gridcellcount)
    else:
        dfmodel = pd.read_csv(
            filename, delim_whitespace=True, header=None,
            skiprows=lambda x: x < headerrows or (x - headerrows - 1) % 2 == 0, names=columns[:5],
            nrows=gridcellcount)
        dfmodeloddlines = pd.read_csv(
            filename, delim_whitespace=True, header=None,
            skiprows=lambda x: x < headerrows or (x - headerrows - 1) % 2 == 1, names=columns[5:],
            nrows=gridcellcount)
        assert len(dfmodel) == len(dfmodeloddlines)
        dfmodel = dfmodel.merge(dfmodeloddlines, left_index=True, right_index=True)
        del dfmodeloddlines

    if len(dfmodel) > gridcellcount:
        dfmodel = dfmodel.iloc[:gridcellcount]

    assert len(dfmodel) == gridcellcount

    dfmodel.index.name = 'cellid'
    # dfmodel.drop('inputcellid', axis=1, inplace=True)

    if dimensions == 1:
        dfmodel['velocity_inner'] = np.concatenate([[0.], dfmodel['velocity_outer'].values[:-1]])
        dfmodel.eval(
            'cellmass_grams = 10 ** logrho * 4. / 3. * 3.14159265 * (velocity_outer ** 3 - velocity_inner ** 3)'
            '* (1e5 * @t_model_init_seconds) ** 3', inplace=True)
        vmax_cmps = dfmodel.velocity_outer.max() * 1e5

    elif dimensions == 3:
        wid_init = artistools.get_wid_init_at_tmodel(modelpath, gridcellcount, t_model_init_days, xmax_tmodel)
        dfmodel.eval('cellmass_grams = rho * @wid_init ** 3', inplace=True)

        dfmodel.rename(columns={
            'pos_x_min': 'pos_x_min', 'pos_y_min': 'pos_y_min', 'pos_z_min': 'pos_z_min'
        }, inplace=True)
        if 'pos_x_min' in dfmodel.columns:
            print("Cell positions in model.txt are defined in the header")
        else:
            cellid = dfmodel.index.values
            xindex = cellid % ncoordgridx
            yindex = (cellid // ncoordgridx) % ncoordgridy
            zindex = (cellid // (ncoordgridx * ncoordgridy)) % ncoordgridz
            dfmodel['pos_x_min'] = -xmax_tmodel + 2 * xindex * xmax_tmodel / ncoordgridx
            dfmodel['pos_y_min'] = -xmax_tmodel + 2 * yindex * xmax_tmodel / ncoordgridy
            dfmodel['pos_z_min'] = -xmax_tmodel + 2 * zindex * xmax_tmodel / ncoordgridz

            def vectormatch(vec1, vec2):
                xclose = np.isclose(vec1[0], vec2[0], atol=xmax_tmodel / ncoordgridx)
                yclose = np.isclose(vec1[1], vec2[1], atol=xmax_tmodel / ncoordgridy)
                zclose = np.isclose(vec1[2], vec2[2], atol=xmax_tmodel / ncoordgridz)

                return all([xclose, yclose, zclose])

            posmatch_xyz = True
            posmatch_zyx = True
            # important cell numbers to check for coordinate column order
            indexlist = [0, ncoordgridx - 1, (ncoordgridx - 1) * (ncoordgridy - 1),
                         (ncoordgridx - 1) * (ncoordgridy - 1) * (ncoordgridz - 1)]
            for index in indexlist:
                cell = dfmodel.iloc[index]
                if not vectormatch([cell.inputpos_a, cell.inputpos_b, cell.inputpos_c],
                                   [cell.pos_x_min, cell.pos_y_min, cell.pos_z_min]):
                    posmatch_xyz = False
                if not vectormatch([cell.inputpos_a, cell.inputpos_b, cell.inputpos_c],
                                   [cell.pos_z_min, cell.pos_y_min, cell.pos_x_min]):
                    posmatch_zyx = False

            assert posmatch_xyz != posmatch_zyx  # one option must match
            if posmatch_xyz:
                print("Cell positions in model.txt are consistent with calculated values when x-y-z column order")
            if posmatch_zyx:
                print("Cell positions in model.txt are consistent with calculated values when z-y-x column order")

    if get_abundances:
        if dimensions == 3:
            print('Getting abundances')
        abundancedata = get_initialabundances(modelpath)
        dfmodel = dfmodel.merge(abundancedata, how='inner', on='inputcellid')

    if derived_cols:
        add_derived_cols_to_modeldata(dfmodel, derived_cols, dimensions, t_model_init_seconds, wid_init, modelpath)

    return dfmodel, t_model_init_days, vmax_cmps


def add_derived_cols_to_modeldata(dfmodel, derived_cols, dimensions=None, t_model_init_seconds=None, wid_init=None,
                                  modelpath=None):
    """add columns to modeldata using e.g. derived_cols = ('velocity', 'Ye')"""
    if dimensions is None:
        dimensions = get_dfmodel_dimensions(dfmodel)

    if dimensions == 3 and 'velocity' in derived_cols:
        dfmodel['vel_x_min'] = dfmodel['pos_x_min'] / t_model_init_seconds
        dfmodel['vel_y_min'] = dfmodel['pos_y_min'] / t_model_init_seconds
        dfmodel['vel_z_min'] = dfmodel['pos_z_min'] / t_model_init_seconds

        dfmodel['vel_x_max'] = (dfmodel['pos_x_min'] + wid_init) / t_model_init_seconds
        dfmodel['vel_y_max'] = (dfmodel['pos_y_min'] + wid_init) / t_model_init_seconds
        dfmodel['vel_z_max'] = (dfmodel['pos_z_min'] + wid_init) / t_model_init_seconds

        dfmodel['vel_x_mid'] = (dfmodel['pos_x_min'] + (0.5 * wid_init)) / t_model_init_seconds
        dfmodel['vel_y_mid'] = (dfmodel['pos_y_min'] + (0.5 * wid_init)) / t_model_init_seconds
        dfmodel['vel_z_mid'] = (dfmodel['pos_z_min'] + (0.5 * wid_init)) / t_model_init_seconds

        dfmodel.eval('vel_mid_radial = sqrt(vel_x_mid ** 2 + vel_y_mid ** 2 + vel_z_mid ** 2)', inplace=True)

    if dimensions == 3 and 'pos_mid' in derived_cols or 'angle_bin' in derived_cols:
        dfmodel['pos_x_mid'] = (dfmodel['pos_x_min'] + (0.5 * wid_init))
        dfmodel['pos_y_mid'] = (dfmodel['pos_y_min'] + (0.5 * wid_init))
        dfmodel['pos_z_mid'] = (dfmodel['pos_z_min'] + (0.5 * wid_init))

    if 'angle_bin' in derived_cols:
        get_cell_angle(dfmodel, modelpath)

    if 'Ye' in derived_cols and os.path.isfile(modelpath / 'Ye.txt'):
        dfmodel['Ye'] = artistools.inputmodel.opacityinputfile.get_Ye_from_file(modelpath)
    if 'Q' in derived_cols and os.path.isfile(modelpath / 'Q_energy.txt'):
        dfmodel['Q'] = artistools.inputmodel.energyinputfiles.get_Q_energy_from_file(modelpath)

    return dfmodel


def get_cell_angle(dfmodel, modelpath):
    """get angle between cell midpoint and axis"""
    syn_dir = artistools.get_syn_dir(modelpath)

    cos_theta = np.zeros(len(dfmodel))
    i = 0
    for _, cell in dfmodel.iterrows():
        mid_point = [cell['pos_x_mid'], cell['pos_y_mid'], cell['pos_z_mid']]
        cos_theta[i] = (
            artistools.dot(mid_point, syn_dir)) / (artistools.vec_len(mid_point) * artistools.vec_len(syn_dir))
        i += 1
    dfmodel['cos_theta'] = cos_theta
    cos_bins = [-1, -0.8, -0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 1]  # including end bin
    labels = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]  # to agree with escaping packet bin numbers
    dfmodel['cos_bin'] = pd.cut(dfmodel['cos_theta'], cos_bins, labels=labels)
    # dfmodel['cos_bin'] = np.searchsorted(cos_bins, dfmodel['cos_theta'].values) -1

    return dfmodel


def get_mean_cell_properties_of_angle_bin(dfmodeldata, vmax_cmps, modelpath=None):
    if 'cos_bin' not in dfmodeldata:
        get_cell_angle(dfmodeldata, modelpath)

    dfmodeldata['rho'][dfmodeldata['rho'] == 0] = None
    dfmodeldata['rho']

    cell_velocities = np.unique(dfmodeldata['vel_x_min'].values)
    cell_velocities = cell_velocities[cell_velocities >= 0]
    velocity_bins = np.append(cell_velocities, vmax_cmps)

    mid_velocities = np.unique(dfmodeldata['vel_x_mid'].values)
    mid_velocities = mid_velocities[mid_velocities >= 0]

    mean_bin_properties = {}
    for bin_number in range(10):
        mean_bin_properties[bin_number] = pd.DataFrame({'velocity': mid_velocities,
                                                        'mean_rho': np.zeros_like(mid_velocities, dtype=float),
                                                        'mean_Ye': np.zeros_like(mid_velocities, dtype=float),
                                                        'mean_Q': np.zeros_like(mid_velocities, dtype=float)})

    # cos_bin_number = 90
    for bin_number in range(10):
        cos_bin_number = bin_number * 10
        # get cells with bin number
        dfanglebin = dfmodeldata.query('cos_bin == @cos_bin_number', inplace=False)

        binned = pd.cut(dfanglebin['vel_mid_radial'], velocity_bins, labels=False, include_lowest=True)
        i = 0
        for binindex, mean_rho in dfanglebin.groupby(binned)['rho'].mean().iteritems():
            i += 1
            mean_bin_properties[bin_number]['mean_rho'][binindex] += mean_rho
        i = 0
        if 'Ye' in dfmodeldata.keys():
            for binindex, mean_Ye in dfanglebin.groupby(binned)['Ye'].mean().iteritems():
                i += 1
                mean_bin_properties[bin_number]['mean_Ye'][binindex] += mean_Ye
        if 'Q' in dfmodeldata.keys():
            for binindex, mean_Q in dfanglebin.groupby(binned)['Q'].mean().iteritems():
                i += 1
                mean_bin_properties[bin_number]['mean_Q'][binindex] += mean_Q

    return mean_bin_properties


def get_2d_modeldata(modelpath):
    filepath = os.path.join(modelpath, 'model.txt')
    num_lines = sum(1 for line in open(filepath))
    skiprowlist = [0, 1, 2]
    skiprowlistodds = skiprowlist + [i for i in range(3, num_lines) if i % 2 == 1]
    skiprowlistevens = skiprowlist + [i for i in range(3, num_lines) if i % 2 == 0]

    model1stlines = pd.read_csv(filepath, delim_whitespace=True, header=None, skiprows=skiprowlistevens)
    model2ndlines = pd.read_csv(filepath, delim_whitespace=True, header=None, skiprows=skiprowlistodds)

    model = pd.concat([model1stlines, model2ndlines], axis=1)
    column_names = ['inputcellid', 'cellpos_mid[r]', 'cellpos_mid[z]', 'rho_model',
                    'X_Fegroup', 'X_Ni56', 'X_Co56', 'X_Fe52', 'X_Cr48']
    model.columns = column_names
    return model


def get_3d_model_data_merged_model_and_abundances_minimal(args):
    """Get 3D data without generating all the extra columns in standard routine.
    Needed for large (eg. 200^3) models"""
    model = get_3d_modeldata_minimal(args.modelpath)
    abundances = get_initialabundances(args.modelpath[0])

    with open(os.path.join(args.modelpath[0], 'model.txt'), 'r') as fmodelin:
        fmodelin.readline()  # npts_model3d
        args.t_model = float(fmodelin.readline())  # days
        args.vmax = float(fmodelin.readline())  # v_max in [cm/s]

    print(model.keys())

    merge_dfs = model.merge(abundances, how='inner', on='inputcellid')

    del model
    del abundances
    gc.collect()

    merge_dfs.info(verbose=False, memory_usage="deep")

    return merge_dfs


def get_3d_modeldata_minimal(modelpath):
    """Read 3D model without generating all the extra columns in standard routine.
    Needed for large (eg. 200^3) models"""
    model = pd.read_csv(os.path.join(modelpath[0], 'model.txt'),
                        delim_whitespace=True, header=None, skiprows=3, dtype=np.float64)
    columns = ['inputcellid', 'cellpos_in[z]', 'cellpos_in[y]', 'cellpos_in[x]', 'rho_model',
               'X_Fegroup', 'X_Ni56', 'X_Co56', 'X_Fe52', 'X_Cr48']
    model = pd.DataFrame(model.values.reshape(-1, 10))
    model.columns = columns

    print('model.txt memory usage:')
    model.info(verbose=False, memory_usage="deep")
    return model


def save_modeldata(
        dfmodel, t_model_init_days, filename=None, modelpath=None, vmax=None, dimensions=1, radioactives=True):
    """Save a pandas DataFrame and snapshot time into ARTIS model.txt"""

    timestart = time.perf_counter()
    assert dimensions in [1, 3, None]
    if dimensions == 1:
        standardcols = ['inputcellid', 'velocity_outer', 'logrho', 'X_Fegroup', 'X_Ni56', 'X_Co56', 'X_Fe52',
                        'X_Cr48']
    elif dimensions == 3:
        dfmodel.rename(columns={'gridindex': 'inputcellid'}, inplace=True)
        griddimension = int(round(len(dfmodel) ** (1. / 3.)))
        print(f' grid size: {len(dfmodel)} ({griddimension}^3)')
        assert griddimension ** 3 == len(dfmodel)

        standardcols = [
            'inputcellid', 'pos_x_min', 'pos_y_min', 'pos_z_min', 'rho',
            'X_Fegroup', 'X_Ni56', 'X_Co56', 'X_Fe52', 'X_Cr48']

    # these two columns are optional, but position is important and they must appear before any other custom cols
    if 'X_Ni57' in dfmodel.columns:
        standardcols.append('X_Ni57')

    if 'X_Co57' in dfmodel.columns:
        standardcols.append('X_Co57')

    dfmodel['inputcellid'] = dfmodel['inputcellid'].astype(int)
    customcols = [col for col in dfmodel.columns if col not in standardcols and col.startswith('X_')]
    customcols.sort(key=lambda col: artistools.get_z_a_nucname(col))  # sort columns by atomic number, mass number

    # set missing radioabundance columns to zero
    for col in standardcols:
        if col not in dfmodel.columns and col.startswith('X_'):
            dfmodel[col] = 0.0

    assert modelpath is not None or filename is not None
    if filename is None:
        filename = 'model.txt'
    if modelpath is not None:
        modelfilepath = Path(modelpath, filename)
    else:
        modelfilepath = Path(filename)

    with open(modelfilepath, 'w') as fmodel:
        fmodel.write(f'{len(dfmodel)}\n')
        fmodel.write(f'{t_model_init_days}\n')
        if dimensions == 3:
            fmodel.write(f'{vmax}\n')

        if customcols:
            fmodel.write(f'#{"  ".join(standardcols)} {"  ".join(customcols)}\n')

        abundcols = [*[col for col in standardcols if col.startswith('X_')], *customcols]

        # for cell in dfmodel.itertuples():
        #     if dimensions == 1:
        #         fmodel.write(f'{cell.inputcellid:6d}   {cell.velocity_outer:9.2f}   {cell.logrho:10.8f} ')
        #     elif dimensions == 3:
        #         fmodel.write(f"{cell.inputcellid:6d} {cell.posx} {cell.posy} {cell.posz} {cell.rho}\n")
        #
        #     fmodel.write(" ".join([f'{getattr(cell, col)}' for col in abundcols]))
        #
        #     fmodel.write('\n')
        if dimensions == 1:

            for cell in dfmodel.itertuples(index=False):
                fmodel.write(f'{cell.inputcellid:6d}   {cell.velocity_outer:9.2f}   {cell.logrho:10.8f} ')
                fmodel.write(" ".join([f'{getattr(cell, col)}' for col in abundcols]))
                fmodel.write('\n')

        elif dimensions == 3:
            zeroabund = ' '.join(['0.0' for _ in abundcols])

            for inputcellid, posxmin, posymin, poszmin, rho, *massfracs in dfmodel[
                    ['inputcellid', 'pos_x_min', 'pos_y_min', 'pos_z_min', 'rho', *abundcols]
                    ].itertuples(index=False, name=None):

                fmodel.write(f"{inputcellid:6d} {posxmin} {posymin} {poszmin} {rho}\n")
                fmodel.write(" ".join([f'{abund}' for abund in massfracs]) if rho > 0. else zeroabund)
                fmodel.write('\n')

    print(f'Saved {filename} (took {time.perf_counter() - timestart:.1f} seconds)')


def get_mgi_of_velocity_kms(modelpath, velocity, mgilist=None):
    """Return the modelgridindex of the cell whose outer velocity is closest to velocity.
    If mgilist is given, then chose from these cells only"""
    modeldata, _, _ = get_modeldata(modelpath)

    velocity = float(velocity)

    if not mgilist:
        mgilist = [mgi for mgi in modeldata.index]
        arr_vouter = modeldata['velocity_outer'].values
    else:
        arr_vouter = np.array([modeldata['velocity_outer'][mgi] for mgi in mgilist])

    index_closestvouter = np.abs(arr_vouter - velocity).argmin()

    if velocity < arr_vouter[index_closestvouter] or index_closestvouter + 1 >= len(mgilist):
        return mgilist[index_closestvouter]
    elif velocity < arr_vouter[index_closestvouter + 1]:
        return mgilist[index_closestvouter + 1]
    elif np.isnan(velocity):
        return float('nan')
    else:
        print(f"Can't find cell with velocity of {velocity}. Velocity list: {arr_vouter}")
        assert(False)


@lru_cache(maxsize=8)
def get_initialabundances(modelpath):
    """Return a list of mass fractions."""
    abundancefilepath = artistools.firstexisting(
        ['abundances.txt.xz', 'abundances.txt.gz', 'abundances.txt'], path=modelpath)

    abundancedata = pd.read_csv(abundancefilepath, delim_whitespace=True, header=None)
    abundancedata.index.name = 'modelgridindex'
    abundancedata.columns = [
        'inputcellid', *['X_' + artistools.get_elsymbol(x) for x in range(1, len(abundancedata.columns))]]
    if len(abundancedata) > 100000:
        print('abundancedata memory usage:')
        abundancedata.info(verbose=False, memory_usage="deep")
    return abundancedata


def save_initialabundances(dfelabundances, abundancefilename):
    """Save a DataFrame (same format as get_initialabundances) to abundances.txt.
        columns must be:
            - inputcellid: integer index to match model.txt (starting from 1)
            - X_El: mass fraction of element with two-letter code 'El' (e.g., X_H, X_He, H_Li, ...)
    """
    timestart = time.perf_counter()
    if Path(abundancefilename).is_dir():
        abundancefilename = Path(abundancefilename) / 'abundances.txt'
    dfelabundances['inputcellid'] = dfelabundances['inputcellid'].astype(int)
    atomic_numbers = [artistools.get_atomic_number(colname[2:])
                      for colname in dfelabundances.columns if colname.startswith('X_')]
    elcolnames = [f'X_{artistools.get_elsymbol(Z)}' for Z in range(1, 1 + max(atomic_numbers))]

    # set missing elemental abundance columns to zero
    for col in elcolnames:
        if col not in dfelabundances.columns:
            dfelabundances[col] = 0.0

    with open(abundancefilename, 'w') as fabund:
        for row in dfelabundances.itertuples(index=False):
            fabund.write(f' {row.inputcellid:6d} ')
            fabund.write(" ".join([f'{getattr(row, colname, 0.)}' for colname in elcolnames]))
            fabund.write("\n")

    print(f'Saved {abundancefilename} (took {time.perf_counter() - timestart:.1f} seconds)')


def save_empty_abundance_file(ngrid, outputfilepath='.'):
    """Dummy abundance file with only zeros"""
    Z_atomic = np.arange(1, 31)

    abundancedata = {'cellid': range(1, ngrid + 1)}
    for atomic_number in Z_atomic:
        abundancedata[f'Z={atomic_number}'] = np.zeros(ngrid)

    # abundancedata['Z=28'] = np.ones(ngrid)

    abundancedata = pd.DataFrame(data=abundancedata)
    abundancedata = abundancedata.round(decimals=5)
    abundancedata.to_csv(Path(outputfilepath) / 'abundances.txt', header=False, sep='\t', index=False)


def get_dfmodel_dimensions(dfmodel):
    if 'pos_x_min' in dfmodel.columns:
        return 3

    return 1


def sphericalaverage(dfmodel, t_model_init_days, vmax, dfelabundances=None, dfgridcontributions=None):
    """Convert 3D Cartesian grid model to 1D spherical"""
    t_model_init_seconds = t_model_init_days * 24 * 60 * 60
    xmax = vmax * t_model_init_seconds
    ngridpoints = len(dfmodel)
    ncoordgridx = round(ngridpoints ** (1. / 3.))
    wid_init = 2 * xmax / ncoordgridx

    print(f'Spherically averaging 3D model with {ngridpoints} cells...')
    timestart = time.perf_counter()

    # dfmodel = dfmodel.query('rho > 0.').copy()
    dfmodel = dfmodel.copy()
    celldensity = {cellindex: rho for cellindex, rho in dfmodel[['inputcellid', 'rho']].itertuples(index=False)}

    dfmodel = add_derived_cols_to_modeldata(
        dfmodel, ['velocity'], dimensions=3, t_model_init_seconds=t_model_init_seconds, wid_init=wid_init)
    # print(dfmodel)
    # print(dfelabundances)
    km_to_cm = 1e5
    velocity_bins = [vmax * n / ncoordgridx for n in range(ncoordgridx + 1)]  # cm/s
    outcells = []
    outcellabundances = []
    outgridcontributions = []

    # cellidmap_3d_to_1d = {}
    highest_active_radialcellid = -1
    for radialcellid, (velocity_inner, velocity_outer) in enumerate(zip(velocity_bins[:-1], velocity_bins[1:]), 1):
        assert velocity_outer > velocity_inner
        matchedcells = dfmodel.query(
            'vel_mid_radial > @velocity_inner and vel_mid_radial <= @velocity_outer')
        matchedcellrhosum = matchedcells.rho.sum()
        # cellidmap_3d_to_1d.update({cellid_3d: radialcellid for cellid_3d in matchedcells.inputcellid})

        if len(matchedcells) == 0:
            rhomean = 0.
        else:
            shell_volume = (4 * math.pi / 3) * (
                (velocity_outer * t_model_init_seconds) ** 3 - (velocity_inner * t_model_init_seconds) ** 3)
            rhomean = matchedcellrhosum * wid_init ** 3 / shell_volume
            # volumecorrection = len(matchedcells) * wid_init ** 3 / shell_volume
            # print(radialcellid, volumecorrection)

            if rhomean > 0. and dfgridcontributions is not None:
                dfcellcont = dfgridcontributions.query('cellindex in @matchedcells.inputcellid.values')

                for particleid, dfparticlecontribs in dfcellcont.groupby('particleid'):
                    frac_of_cellmass_avg = sum([
                        (row.frac_of_cellmass *
                         celldensity[row.cellindex])
                        for row in dfparticlecontribs.itertuples(index=False)]) / matchedcellrhosum

                    frac_of_cellmass_includemissing_avg = sum([
                        (row.frac_of_cellmass_includemissing *
                         celldensity[row.cellindex])
                        for row in dfparticlecontribs.itertuples(index=False)]) / matchedcellrhosum

                    outgridcontributions.append({
                        'particleid': particleid,
                        'cellindex': radialcellid,
                        'frac_of_cellmass': frac_of_cellmass_avg,
                        'frac_of_cellmass_includemissing': frac_of_cellmass_includemissing_avg,
                    })

        if rhomean > 0.:
            highest_active_radialcellid = radialcellid
        logrho = math.log10(max(1e-99, rhomean))

        dictcell = {
            'inputcellid': radialcellid,
            'velocity_outer': velocity_outer / km_to_cm,
            'logrho': logrho,
        }

        for column in matchedcells.columns:
            if column.startswith('X_'):
                if rhomean > 0.:
                    massfrac = np.dot(matchedcells[column], matchedcells.rho) / matchedcellrhosum
                else:
                    massfrac = 0.
                dictcell[column] = massfrac

        outcells.append(dictcell)

        if dfelabundances is not None:
            if rhomean > 0.:
                abund_matchedcells = dfelabundances.loc[matchedcells.index]
            else:
                abund_matchedcells = None
            dictcellabundances = {'inputcellid': radialcellid}
            for column in dfelabundances.columns:
                if column.startswith('X_'):
                    if rhomean > 0.:
                        massfrac = np.dot(abund_matchedcells[column], matchedcells.rho) / matchedcellrhosum
                    else:
                        massfrac = 0.
                    dictcellabundances[column] = massfrac

            outcellabundances.append(dictcellabundances)

    dfmodel1d = pd.DataFrame(outcells[:highest_active_radialcellid])

    dfabundances1d = (
        pd.DataFrame(outcellabundances[:highest_active_radialcellid]) if outcellabundances else None)

    dfgridcontributions1d = pd.DataFrame(outgridcontributions) if outgridcontributions else None
    print(f'  took {time.perf_counter() - timestart:.1f} seconds')

    return dfmodel1d, dfabundances1d, dfgridcontributions1d



#!/usr/bin/env python3

# import hashlib
# import math
# import numpy as np
# import os
# import os.path
import pandas as pd
# import pytest
from pathlib import Path

import artistools as at
import artistools.inputmodel

modelpath = at.config['path_testartismodel']
outputpath = Path(at.config['path_testoutput'])
# outputpath.mkdir(parents=True, exist_ok=True)
at.enable_diskcache = False


def test_describeinputmodel():
    at.inputmodel.describeinputmodel.main(argsraw=[], inputfile=modelpath, get_abundances=True)


def test_makemodel_botyanski2017():
    at.inputmodel.botyanski2017.main(argsraw=[], outputpath=outputpath)


def test_makemodel():
    at.inputmodel.makeartismodel.main(argsraw=[], modelpath=modelpath, outputpath=outputpath)


def test_makemodel_energyfiles():
    at.inputmodel.makeartismodel.main(
        argsraw=[], modelpath=modelpath, makeenergyinputfiles=True, modeldim=1, outputpath=outputpath)


def test_maketardismodel():
    at.inputmodel.maketardismodelfromartis.main(
        argsraw=[], inputpath=modelpath, outputpath=outputpath)


def test_make_empty_abundance_file():
    at.inputmodel.save_empty_abundance_file(ngrid=50, outputfilepath=outputpath)


def test_opacity_by_Ye_file():
    griddata = {'cellYe': [0, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.5],
                'rho': [0, 99, 99, 99, 99, 99, 99, 99],
                'inputcellid': range(1, 9)}
    at.inputmodel.opacityinputfile.opacity_by_Ye(outputpath, griddata=griddata)


def test_save3Dmodel():
    dfmodel = pd.DataFrame(
        {'inputcellid': [1, 2, 3, 4, 5, 6, 7, 8],
         'pos_x_min': [-1, 1, -1, 1, -1, 1, -1, 1],
         'pos_y_min': [-1, -1, 1, 1, -1, -1, 1, 1],
         'pos_z_min': [-1, -1, -1, -1, 1, 1, 1, 1],
         'rho': [0, 2, 3, 2, 5, 7, 8, 2],
         'cellYe': [0, 0.1, 0.2, .1, .5, .1, .3, 3]})
    tmodel = 100
    vmax = 1000
    at.inputmodel.save_modeldata(
        modelpath=outputpath, dfmodel=dfmodel, t_model_init_days=tmodel,
        vmax=vmax, dimensions=3)



#!/usr/bin/env python3

import argparse
import numpy as np
from pathlib import Path

import pandas as pd

import artistools as at
import artistools.estimators
import artistools.lightcurve


def write_spectra(modelpath, model_id, selected_timesteps, outfile):
    spec_data = np.loadtxt(Path(modelpath, "spec.out"))

    times = spec_data[0, 1:]
    freqs = spec_data[1:, 0]
    lambdas = 2.99792458e18 / freqs

    # print("\n".join(["{0}, {1}".format(*x) for x in enumerate(times)]))

    fluxes_nu = spec_data[1:, 1:]

    # 1 parsec in cm is 3.086e18
    # area in cm^2 of a spherical of radius 1 Mpc is:
    area = 3.086e18 * 3.086e18 * 1e12 * 4. * 3.141592654

    lum_lambda = np.zeros((len(lambdas), len(times)))

    # convert flux to power by multiplying by area
    for n in range(1000):
        # 2.99792458e18 is c in Angstrom / second
        lum_lambda[n, :] = fluxes_nu[n, :] * 2.99792458e18 / lambdas[n] / lambdas[n] * area

    with open(outfile, "w") as f:
        f.write("#NTIMES: {0}\n".format(len(selected_timesteps)))
        f.write("#NWAVE: {0}\n".format(len(lambdas)))
        f.write("#TIMES[d]: {0}\n".format(" ".join(["{0:.2f}".format(times[ts]) for ts in selected_timesteps])))
        f.write("#wavelength[Ang] flux_t0[erg/s/Ang] flux_t1[erg/s/Ang] ... flux_tn[erg/s/Ang]\n")

        for n in reversed(range(len(lambdas))):
            f.write("{0:.2f} ".format(lambdas[n]) + " ".join(
                ["{0:.2e}".format(lum_lambda[n, ts]) for ts in selected_timesteps]) + "\n")

        f.close()


def write_ntimes_nvel(f, selected_timesteps, modelpath):
    times = at.get_timestep_times_float(modelpath)
    modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata(modelpath)
    f.write(f'#NTIMES: {len(selected_timesteps)}\n')
    f.write(f'#NVEL: {len(modeldata)}\n')
    f.write(f'#TIMES[d]: {" ".join([f"{times[ts]:.2f}" for ts in selected_timesteps])}\n')


def write_single_estimator(modelpath, selected_timesteps, estimators, allnonemptymgilist, outfile, keyname):
    modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata(modelpath)
    with open(outfile, "w") as f:
        write_ntimes_nvel(f, selected_timesteps, modelpath)
        if keyname == 'total_dep':
            f.write('#vel_mid[km/s] Edep_t0[erg/s/cm^3] Edep_t1[erg/s/cm^3] ... Edep_tn[erg/s/cm^3]\n')
        elif keyname == 'nne':
            f.write('#vel_mid[km/s] ne_t0[/cm^3] ne_t1[/cm^3] … ne_tn[/cm^3]\n')
        elif keyname == 'Te':
            f.write('#vel_mid[km/s] Tgas_t0[K] Tgas_t1[K] ... Tgas_tn[K]\n')
        for modelgridindex, cell in modeldata.iterrows():
            if modelgridindex not in allnonemptymgilist:
                continue
            v_mid = (cell.velocity_inner + cell.velocity_outer) / 2.
            f.write(f'{v_mid:.2f}')
            for timestep in selected_timesteps:
                cellvalue = estimators[(timestep, modelgridindex)][keyname]
                # try:
                #     cellvalue = estimators[(timestep, modelgridindex)][keyname]
                # except KeyError:
                #     cellvalue = (estimators[(timestep - 1, modelgridindex)][keyname]
                #                  + estimators[(timestep + 1, modelgridindex)][keyname]) / 2.
                f.write(f' {cellvalue:.3e}')
            f.write('\n')


def write_ionfracts(modelpath, model_id, selected_timesteps, estimators, allnonemptymgilist, outputpath):
    times = at.get_timestep_times_float(modelpath)
    modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata(modelpath)
    elementlist = at.get_composition_data(modelpath)
    nelements = len(elementlist)
    for element in range(nelements):
        atomic_number = elementlist.Z[element]
        elsymb = at.get_elsymbol(atomic_number).lower()
        nions = elementlist.nions[element]
        pathfileout = Path(outputpath, f'ionfrac_{elsymb}_{model_id}_artisnebular.txt')
        fileisallzeros = True  # will be changed when a non-zero is encountered
        with open(pathfileout, 'w') as f:
            f.write(f'#NTIMES: {len(selected_timesteps)}\n')
            f.write(f'#NSTAGES: {nions}\n')
            f.write(f'#TIMES[d]: {" ".join([f"{times[ts]:.2f}" for ts in selected_timesteps])}\n')
            f.write('#\n')
            for timestep in selected_timesteps:
                f.write(f'#TIME: {times[timestep]:.2f}\n')
                f.write(f'#NVEL: {len(allnonemptymgilist)}\n')
                f.write(f'#vel_mid[km/s] {" ".join([f"{elsymb}{ion}" for ion in range(nions)])}\n')
                for modelgridindex, cell in modeldata.iterrows():
                    if modelgridindex not in allnonemptymgilist:
                        continue
                    v_mid = (cell.velocity_inner + cell.velocity_outer) / 2.
                    f.write(f'{v_mid:.2f}')
                    elabund = estimators[(timestep, modelgridindex)]['populations'].get(atomic_number, 0)
                    for ion in range(nions):
                        ion_stage = ion + elementlist.lowermost_ionstage[element]
                        ionabund = estimators[(timestep, modelgridindex)]['populations'].get(
                            (atomic_number, ion_stage), 0)
                        ionfrac = ionabund / elabund if elabund > 0 else 0
                        if ionfrac > 0.:
                            fileisallzeros = False
                        f.write(' {:.4e}'.format(ionfrac))
                    f.write('\n')
        if fileisallzeros:
            print(f'Deleting {pathfileout} because it is all zeros')
            pathfileout.unlink()


def write_phys(modelpath, model_id, selected_timesteps, estimators, allnonemptymgilist, outputpath):
    times = at.get_timestep_times_float(modelpath)
    modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata(modelpath)
    with open(Path(outputpath, f'phys_{model_id}_artisnebular.txt'), 'w') as f:
        f.write(f'#NTIMES: {len(selected_timesteps)}\n')
        f.write(f'#TIMES[d]: {" ".join([f"{times[ts]:.2f}" for ts in selected_timesteps])}\n')
        f.write('#\n')
        for timestep in selected_timesteps:
            f.write(f'#TIME: {times[timestep]:.2f}\n')
            f.write(f'#NVEL: {len(modeldata)}\n')
            f.write('#vel_mid[km/s] temp[K] rho[gcc] ne[/cm^3] natom[/cm^3]\n')
            for modelgridindex, cell in modeldata.iterrows():
                if modelgridindex not in allnonemptymgilist:
                    continue

                estimators[(timestep, modelgridindex)]['rho'] = (
                    10 ** cell.logrho * (t_model_init_days / times[timestep]) ** 3)

                estimators[(timestep, modelgridindex)]['nntot'] = (
                    estimators[(timestep, modelgridindex)]['populations']['total'])

                v_mid = (cell.velocity_inner + cell.velocity_outer) / 2.
                f.write(f'{v_mid:.2f}')
                for keyname in ('Te', 'rho', 'nne', 'nntot'):
                    estvalue = estimators[(timestep, modelgridindex)][keyname]
                    f.write(' {:.4e}'.format(estvalue))
                f.write('\n')


def write_lbol_edep(modelpath, model_id, selected_timesteps, estimators, outputpath):
    # times = at.get_timestep_times_float(modelpath)
    dflightcurve = at.lightcurve.readfile(Path(modelpath, "light_curve.out"))
    dfdep = at.get_deposition(modelpath)

    df = pd.merge(dflightcurve, dfdep, left_index=True, right_index=True, suffixes=('', '_dep'))

    with open(outputpath, "w") as f:
        f.write(f'#NTIMES: {len(selected_timesteps)}\n')
        f.write("#time[d] Lbol[erg/s] Edep[erg/s] \n")

        for timestep, row in df.iterrows():
            if timestep not in selected_timesteps:
                continue
            f.write(f"{row.time:.2f} {row.lum * 3.826e33:.2e} {row.total_dep_Lsun * 3.826e33:.2e}\n")

    f.close()


def addargs(parser):
    parser.add_argument('-modelpath', default=[], nargs='*', action=at.AppendPath,
                        help='Paths to ARTIS folders')

    parser.add_argument('-selected_timesteps', default=[], nargs='*', action=at.AppendPath,
                        help='Selected ARTIS timesteps')

    parser.add_argument('-outputpath', '-o', action='store', type=Path, default=Path(),
                        help='path for output files')


def main(args=None, argsraw=None, **kwargs):
    """Plot spectra from ARTIS and reference data."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Write ARTIS model data out in code comparison workshop format.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if not args.modelpath and not args.specpath:
        args.modelpath = [Path('.')]
    elif isinstance(args.modelpath, (str, Path)):
        args.modelpath = [args.modelpath]

    modelpathlist = args.modelpath
    selected_timesteps = args.selected_timesteps

    args.outputpath.mkdir(parents=True, exist_ok=True)

    for modelpath in modelpathlist:
        model_id = str(modelpath.name).split('_')[0]

        modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata(modelpath)
        estimators = at.estimators.read_estimators(modelpath=modelpath)
        allnonemptymgilist = [modelgridindex for modelgridindex in modeldata.index
                              if not estimators[(selected_timesteps[0], modelgridindex)]['emptycell']]

        write_lbol_edep(modelpath, model_id, selected_timesteps, estimators,
                        Path(args.outputpath, "lbol_edep_" + model_id + "_artisnebular.txt"))

        write_spectra(modelpath, model_id, selected_timesteps,
                      Path(args.outputpath, "spectra_" + model_id + "_artisnebular.txt"))

        # write_single_estimator(modelpath, selected_timesteps, estimators, allnonemptymgilist,
        #                        Path(args.outputpath, "eden_" + model_id + "_artisnebular.txt"), keyname='nne')

        write_single_estimator(modelpath, selected_timesteps, estimators, allnonemptymgilist,
                               Path(args.outputpath, "edep_" + model_id + "_artisnebular.txt"), keyname='total_dep')

        # write_single_estimator(modelpath, selected_timesteps, estimators, allnonemptymgilist,
        #                        Path(args.outputpath, "tgas_" + model_id + "_artisnebular.txt"), keyname='Te')

        write_phys(modelpath, model_id, selected_timesteps, estimators, allnonemptymgilist, args.outputpath)
        write_ionfracts(modelpath, model_id, selected_timesteps, estimators, allnonemptymgilist, args.outputpath)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import numpy as np

from pathlib import Path
import artistools as at
import artistools.estimators


def addargs(parser):
    parser.add_argument('-outputpath', '-o',
                        default='massfracs.txt',
                        help='Path to output file of mass fractions')


def main(args=None, argsraw=None, **kwargs) -> None:
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Create solar r-process pattern in ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    modelpath = Path('.')
    timestep = 14
    elmass = {el.Z: el.mass for _, el in at.get_composition_data(modelpath).iterrows()}
    outfilename = args.outputpath
    with open(outfilename, 'wt') as fout:
        modelgridindexlist = range(10)
        estimators = at.estimators.read_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindexlist)
        for modelgridindex in modelgridindexlist:
            tdays = estimators[(timestep, modelgridindex)]['tdays']
            popdict = estimators[(timestep, modelgridindex)]['populations']

            numberdens = {}
            totaldens = 0.  # number density times atomic mass summed over all elements
            for key in popdict.keys():
                try:
                    atomic_number = int(key)
                    numberdens[atomic_number] = popdict[atomic_number]
                    totaldens += numberdens[atomic_number] * elmass[atomic_number]
                except ValueError:
                    pass
                except TypeError:
                    pass

            massfracs = {
                atomic_number: numberdens[atomic_number] * elmass[atomic_number] / totaldens
                for atomic_number in numberdens.keys()
            }

            fout.write(f'{tdays}d shell {modelgridindex}\n')
            massfracsum = 0.
            for atomic_number in massfracs.keys():
                massfracsum += massfracs[atomic_number]
                fout.write(f'{atomic_number} {at.get_elsymbol(atomic_number)} {massfracs[atomic_number]}\n')

            assert np.isclose(massfracsum, 1.0)

    print(f'Saved {outfilename}')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import yaml
from pathlib import Path


def main():
    with Path('metadata.yml').open('r') as yamlfile:
        metadata = yaml.load(yamlfile, Loader=yaml.FullLoader)

    for obsfile in metadata.keys():
        metafilepath = Path(obsfile).with_suffix(Path(obsfile).suffix + '.meta.yml')
        with open(metafilepath, 'w') as metafile:
            yaml.dump(metadata[obsfile], metafile)


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
import numpy as np


def set_axis_properties(ax, args):
    if 'subplots' not in args:
        args.subplots = False
    if 'labelfontsize' not in args:
        args.labelfontsize = 18

    if args.subplots:
        for axis in ax:
            axis.minorticks_on()
            axis.tick_params(axis='both', which='minor', top=True, right=True, length=5, width=2,
                             labelsize=args.labelfontsize, direction='in')
            axis.tick_params(axis='both', which='major', top=True, right=True, length=8, width=2,
                             labelsize=args.labelfontsize, direction='in')

    else:
        ax.minorticks_on()
        ax.tick_params(axis='both', which='minor', top=True, right=True, length=5, width=2,
                       labelsize=args.labelfontsize, direction='in')
        ax.tick_params(axis='both', which='major', top=True, right=True, length=8, width=2,
                       labelsize=args.labelfontsize, direction='in')

    if 'ymin' in args or 'ymax' in args:
        plt.ylim(args.ymin, args.ymax)
    if 'xmin' in args or 'xmax' in args:
        plt.xlim(args.xmin, args.xmax)

    plt.minorticks_on()
    return ax


def set_axis_labels(fig, ax, xlabel, ylabel, labelfontsize, args):
    if args.subplots:
        fig.text(0.5, 0.02, xlabel, ha='center', va='center')
        fig.text(0.02, 0.5, ylabel, ha='center', va='center', rotation='vertical')
    else:
        ax.set_xlabel(xlabel, fontsize=labelfontsize)
        ax.set_ylabel(ylabel, fontsize=labelfontsize)


def imshow_init_for_artis_grid(ngrid, vmax, plot_variable_3d_array, plot_axes='xy'):
    # ngrid = round(len(model['inputcellid']) ** (1./3.))
    extent = {'left': -vmax, 'right': vmax, 'bottom': vmax, 'top': -vmax}
    extent = extent['left'], extent['right'], extent['bottom'], extent['top']
    data = np.zeros((ngrid, ngrid))

    plot_axes_choices = ['xy', 'zx']
    if plot_axes not in plot_axes_choices:
        print(f'Choose plot axes from {plot_axes_choices}')
        quit()

    for z in range(0, ngrid):
        for y in range(0, ngrid):
            for x in range(0, ngrid):
                if plot_axes == 'xy':
                    if z == round(ngrid/2)-1:
                        data[x, y] = plot_variable_3d_array[x, y, z]
                elif plot_axes == 'zx':
                    if y == round(ngrid/2)-1:
                        data[z, x] = plot_variable_3d_array[x, y, z]

    return data, extent



#!/usr/bin/env python3
"""Artistools - NLTE population related functions."""
import math
import multiprocessing
# import os
import re
# import sys
from functools import lru_cache
from functools import partial
from pathlib import Path
# from itertools import chain

# import matplotlib.pyplot as plt
# import matplotlib.ticker as ticker
# import numpy as np
import pandas as pd
from astropy import constants as const
# import numpy as np
# import matplotlib as mpl

import artistools as at


def texifyterm(strterm):
    """Replace a term string with TeX notation equivalent."""
    strtermtex = ''
    passed_term_Lchar = False

    for termpiece in re.split('([_A-Za-z])', strterm):
        if re.match('[0-9]', termpiece) is not None and not passed_term_Lchar:
            # 2S + 1 number
            strtermtex += r'$^{' + termpiece + r'}$'
        elif re.match('[A-Z]', termpiece) is not None:
            # L character - SPDFGH...
            strtermtex += termpiece
            passed_term_Lchar = True
        elif re.match('[eo]', termpiece) is not None and passed_term_Lchar:
            # odd flag, but don't want to confuse it with the energy index (e.g. o4Fo[2])
            if termpiece != 'e':  # even is assumed by default (and looks neater with all the 'e's)
                strtermtex += r'$^{\rm ' + termpiece + r'}$'
        elif re.match(r'[0-9]?.*\]', termpiece) is not None:
            # J value
            strtermtex += termpiece.split('[')[0] + r'$_{' + termpiece.lstrip('0123456789').strip('[]') + r'}$'
        elif re.match('[0-9]', termpiece) is not None and passed_term_Lchar:
            # extra number after S char
            strtermtex += termpiece

    strtermtex = strtermtex.replace('$$', '')
    return strtermtex


def texifyconfiguration(levelname):
    """Replace a level configuration with the formatted LaTeX equivalent."""
    # the underscore gets confused with LaTeX subscript operator, so switch it to the hash symbol
    strout = '#'.join(levelname.split('_')[:-1]) + '#'
    for strorbitalocc in re.findall(r'[0-9][a-z][0-9]?[#(]', strout):
        n, lchar, occ = re.split('([a-z])', strorbitalocc)
        lastchar = '(' if occ.endswith('(') else '#'
        occ = occ.rstrip('#(')
        strorbitalocctex = n + lchar + (r'$^{' + occ + r'}$' if occ else '') + lastchar
        strout = strout.replace(strorbitalocc, strorbitalocctex)

    for parentterm in re.findall(r'\([0-9][A-Z][^)]?\)', strout):
        parentermtex = f'({texifyterm(parentterm.strip("()"))})'
        strout = strout.replace(parentterm, parentermtex)
    strterm = levelname.split('_')[-1]
    strout += ' ' + texifyterm(strterm)

    strout = strout.replace('#', '')
    strout = strout.replace('$$', '')
    # print(f"Replacing levelname '{levelname}' with '{strout}'")
    return strout


def add_lte_pops(modelpath, dfpop, columntemperature_tuples, noprint=False, maxlevel=-1):
    """Add columns to dfpop with LTE populations.

    columntemperature_tuples is a sequence of tuples of column name and temperature, e.g., ('mycolumn', 3000)
    """
    k_b = const.k_B.to('eV / K').value

    for _, row in dfpop.drop_duplicates(['modelgridindex', 'timestep', 'Z', 'ion_stage']).iterrows():
        modelgridindex = int(row.modelgridindex)
        timestep = int(row.timestep)
        Z = int(row.Z)
        ion_stage = int(row.ion_stage)

        ionlevels = at.atomic.get_levels(modelpath).query('Z == @Z and ion_stage == @ion_stage').iloc[0].levels

        gs_g = ionlevels.iloc[0].g
        gs_energy = ionlevels.iloc[0].energy_ev

        gs_pop = dfpop.query(
            'modelgridindex == @modelgridindex and timestep == @timestep '
            'and Z == @Z and ion_stage == @ion_stage and level == 0').iloc[0]['n_NLTE']

        masksuperlevel = (
            (dfpop['modelgridindex'] == modelgridindex) & (dfpop['timestep'] == timestep)
            & (dfpop['Z'] == Z) & (dfpop['ion_stage'] == ion_stage) & (dfpop['level'] == -1))

        masknotsuperlevel = (
            (dfpop['modelgridindex'] == modelgridindex) & (dfpop['timestep'] == timestep)
            & (dfpop['Z'] == Z) & (dfpop['ion_stage'] == ion_stage) & (dfpop['level'] != -1))

        def f_ltepop(x, T_exc, gsg, gse, ionlevels):
            return (ionlevels.iloc[int(x.level)].g / gsg
                    * math.exp(- (ionlevels.iloc[int(x.level)].energy_ev - gse) / k_b / T_exc))

        for columnname, T_exc in columntemperature_tuples:
            dfpop.loc[masknotsuperlevel, columnname] = dfpop.loc[masknotsuperlevel].apply(
                f_ltepop, args=(T_exc, gs_g, gs_energy, ionlevels), axis=1)

        if not dfpop[masksuperlevel].empty:
            levelnumber_sl = dfpop.query(
                'modelgridindex == @modelgridindex and timestep == @timestep '
                'and Z == @Z and ion_stage == @ion_stage').level.max() + 1

            if maxlevel < 0 or levelnumber_sl <= maxlevel:
                if not noprint:
                    print(f'{at.get_elsymbol(Z)} {at.roman_numerals[ion_stage]} '
                          f'has a superlevel at level {levelnumber_sl}')

                for columnname, T_exc in columntemperature_tuples:
                    dfpop.loc[masksuperlevel, columnname] = ionlevels.iloc[levelnumber_sl:].eval(
                        'g / @gs_g * exp(- (energy_ev - @gs_energy) / @k_b / @T_exc)').sum()

            dfpop.loc[masksuperlevel, 'level'] = levelnumber_sl + 2

    return dfpop


@at.diskcache(savezipped=True)
def read_file(nltefilepath):
    """Read NLTE populations from one file."""

    if not nltefilepath.is_file():
        nltefilepathgz = Path(str(nltefilepath) + '.gz')
        nltefilepathxz = Path(str(nltefilepath) + '.xz')
        if nltefilepathxz.is_file():
            nltefilepath = nltefilepathxz
        elif nltefilepathgz.is_file():
            nltefilepath = nltefilepathgz
        else:
            # print(f'Warning: Could not find {nltefilepath}')
            return pd.DataFrame()

    filesize = Path(nltefilepath).stat().st_size / 1024 / 1024
    print(f'Reading {nltefilepath} ({filesize:.2f} MiB)')

    try:
        dfpop = pd.read_csv(nltefilepath, delim_whitespace=True)
    except pd.errors.EmptyDataError:
        return pd.DataFrame()

    return dfpop


def read_file_filtered(nltefilepath, strquery=None, dfqueryvars=None):
    dfpopfile = read_file(nltefilepath)

    if strquery and not dfpopfile.empty:
        dfpopfile.query(strquery, local_dict=dfqueryvars, inplace=True)

    return dfpopfile


@lru_cache(maxsize=2)
@at.diskcache(savezipped=True, funcversion="2020-07-03.1327", saveonly=False)
def read_files(modelpath, timestep=-1, modelgridindex=-1, dfquery=None, dfqueryvars={}):
    """Read in NLTE populations from a model for a particular timestep and grid cell."""

    mpiranklist = at.get_mpiranklist(modelpath, modelgridindex=modelgridindex)

    dfpop = pd.DataFrame()

    nltefilepaths = [Path(folderpath, f'nlte_{mpirank:04d}.out')
                     for folderpath in at.get_runfolders(modelpath, timestep=timestep) for mpirank in mpiranklist]

    dfqueryvars['modelgridindex'] = modelgridindex
    dfqueryvars['timestep'] = timestep

    dfquery_full = ''
    if timestep >= 0:
        dfquery_full = 'timestep==@timestep'

    if modelgridindex >= 0:
        if dfquery_full:
            dfquery_full += ' and '
        dfquery_full += 'modelgridindex==@modelgridindex'

    if dfquery:
        if dfquery_full:
            dfquery_full = f'({dfquery_full}) and '
        dfquery_full += f'({dfquery})'

    if at.num_processes > 1:
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            arr_dfnltepop = pool.map(partial(read_file_filtered, strquery=dfquery_full, dfqueryvars=dfqueryvars),
                                     nltefilepaths)
            pool.close()
            pool.join()
            pool.terminate()
    else:
        arr_dfnltepop = [read_file_filtered(f, strquery=dfquery_full, dfqueryvars=dfqueryvars) for f in nltefilepaths]

    dfpop = pd.concat(arr_dfnltepop).copy()

    return dfpop



#!/usr/bin/env python3
import argparse
import math
import os
from pathlib import Path

import artistools as at
import artistools.inputmodel.opacityinputfile
import matplotlib
# import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u

# import artistools.inputmodel
# from mpl_toolkits.mplot3d import Axes3D


def plot_2d_initial_abundances(modelpath, args):
    model = at.inputmodel.get_2d_modeldata(modelpath[0])
    abundances = at.inputmodel.get_initialabundances(modelpath[0])

    abundances['inputcellid'] = abundances['inputcellid'].apply(lambda x: float(x))

    merge_dfs = model.merge(abundances, how='inner', on='inputcellid')

    with open(os.path.join(modelpath[0], 'model.txt'), 'r') as fmodelin:
        fmodelin.readline()  # npts r, npts z
        t_model = float(fmodelin.readline())  # days
        vmax = float(fmodelin.readline())  # v_max in [cm/s]

    r = merge_dfs['cellpos_mid[r]'] / t_model * (u.cm / u.day).to('km/s') / 10 ** 3
    z = merge_dfs['cellpos_mid[z]'] / t_model * (u.cm / u.day).to('km/s') / 10 ** 3

    ion = f'X_{args.ion}'
    font = {'weight': 'bold',
            'size': 18}

    f = plt.figure(figsize=(4, 5))
    ax = f.add_subplot(111)
    im = ax.scatter(r, z, c=merge_dfs[ion], marker="8")

    f.colorbar(im)
    plt.xlabel(r"v$_x$ in 10$^3$ km/s", fontsize='x-large')  # , fontweight='bold')
    plt.ylabel(r"v$_z$ in 10$^3$ km/s", fontsize='x-large')  # , fontweight='bold')
    plt.text(20, 25, args.ion, color='white', fontweight='bold', fontsize='x-large')
    plt.tight_layout()
    # ax.labelsize: 'large'
    # plt.title(f'At {sliceaxis} = {sliceposition}')

    outfilename = f'plotcomposition{args.ion}.pdf'
    plt.savefig(Path(modelpath[0]) / outfilename, format='pdf')
    print(f'Saved {outfilename}')


def get_merged_model_abundances(modelpath):
    # t_model is in days and vmax is in cm/s
    model, t_model, vmax = at.inputmodel.get_modeldata(modelpath[0], dimensions=3)

    abundances = at.inputmodel.get_initialabundances(modelpath[0])

    abundances['inputcellid'] = abundances['inputcellid'].apply(lambda x: float(x))

    merge_dfs = model.merge(abundances, how='inner', on='inputcellid')
    return merge_dfs, t_model


def get_2D_slice_through_3d_model(merge_dfs, sliceaxis, sliceindex=None):
    if not sliceindex:
        # get midpoint
        sliceposition = merge_dfs.iloc[(merge_dfs['pos_x_min']).abs().argsort()][:1]['pos_x_min'].item()
        # Choose position to slice. This gets minimum absolute value as the closest to 0
    else:
        cell_boundaries = []
        [cell_boundaries.append(x) for x in merge_dfs[f'pos_{sliceaxis}'] if x not in cell_boundaries]
        sliceposition = cell_boundaries[sliceindex]

    slicedf = (merge_dfs.loc[merge_dfs[f'pos_{sliceaxis}'] == sliceposition])
    return slicedf


def plot_abundances_ion(ax, plotvals, ion, plotaxis1, plotaxis2, t_model):
    colorscale = plotvals[ion]
    # colorscale = np.ma.masked_where(colorscale == 0., colorscale)
    # colorscale = np.log10(colorscale)

    norm = matplotlib.colors.Normalize(vmin=0, vmax=1)
    scaledmap = matplotlib.cm.ScalarMappable(cmap='viridis', norm=norm)
    scaledmap.set_array([])
    colorscale = scaledmap.to_rgba(colorscale)  # colorscale fixed between 0 and 1

    x = plotvals[f'pos_{plotaxis1}'] / t_model * (u.cm / u.day).to('km/s') / 10 ** 3
    y = plotvals[f'pos_{plotaxis2}'] / t_model * (u.cm / u.day).to('km/s') / 10 ** 3

    im = ax.scatter(x, y, c=colorscale, marker="8", rasterized=True)  # cmap=plt.get_cmap('PuOr')

    ymin, ymax = ax.get_ylim()
    xmin, xmax = ax.get_xlim()
    ax.text(xmax * 0.6, ymax * 0.7, ion.split('_')[1], color='k', fontweight='bold')
    return im, scaledmap


def plot_3d_initial_abundances(modelpath, args=None):
    font = {
        # 'weight': 'bold',
        'size': 18
    }
    matplotlib.rc('font', **font)

    merge_dfs, t_model = get_merged_model_abundances(modelpath)
    # merge_dfs = plot_most_abundant(modelpath, args)

    plotaxis1 = 'y'
    plotaxis2 = 'z'
    sliceaxis = 'x'

    plotvals = get_2D_slice_through_3d_model(merge_dfs, sliceaxis)

    subplots = False
    if len(args.ion) > 1:
        subplots = True

    if not subplots:
        fig = plt.figure(figsize=(8, 7))
        ax = plt.subplot(111, aspect='equal')
    else:
        rows = 1
        cols = len(args.ion)

        fig, axes = plt.subplots(nrows=rows, ncols=cols, sharex=True, sharey=True,
                                 figsize=(at.figwidth * cols,
                                          at.figwidth * 1.4),
                                 tight_layout={"pad": 5.0, "w_pad": 0.0, "h_pad": 0.0})
        for ax in axes:
            ax.set(aspect='equal')

    for index, ion in enumerate(args.ion):
        ion = f'X_{ion}'
        if subplots:
            ax = axes[index]
        im, scaledmap = plot_abundances_ion(ax, plotvals, ion, plotaxis1, plotaxis2, t_model)

    xlabel = fr"v$_{plotaxis1}$ in 10$^3$ km/s"
    ylabel = fr"v$_{plotaxis2}$ in 10$^3$ km/s"

    if not subplots:
        cbar = plt.colorbar(im)
        plt.xlabel(xlabel, fontsize='x-large')  # , fontweight='bold')
        plt.ylabel(ylabel, fontsize='x-large')  # , fontweight='bold')
    else:
        cbar = fig.colorbar(scaledmap, ax=axes, shrink=cols * 0.08, location='top', pad=0.8, anchor=(0.5, 3.))
        fig.text(0.5, 0.15, xlabel, ha='center', va='center')
        fig.text(0.05, 0.5, ylabel, ha='center', va='center', rotation='vertical')

    # cbar.set_label(label=ion, size='x-large') #, fontweight='bold')
    # cbar.ax.set_title(f'{args.ion}', size='small')
    # cbar.ax.tick_params(labelsize='x-large')

    # plt.tight_layout()
    # ax.labelsize: 'large'
    # plt.title(f'At {sliceaxis} = {sliceposition}')

    if args.outputfile:
        outfilename = args. outputfile
    else:
        outfilename = f'plotcomposition{ion}.pdf'
    # plt.savefig(Path(modelpath[0]) / outfilename, format='pdf')
    plt.savefig(outfilename, format='pdf')
    print(f'Saved {outfilename}')


def get_model_abundances_Msun_1D(modelpath):
    filename = modelpath / 'model.txt'
    modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata(filename)
    abundancedata = at.inputmodel.get_initialabundances(modelpath)

    t_model_init_seconds = t_model_init_days * 24 * 60 * 60

    modeldata['volume_shell'] = 4 / 3 * math.pi * ((modeldata['velocity_outer'] * 1e5 * t_model_init_seconds) ** 3
                                                   - (modeldata['velocity_inner'] * 1e5 * t_model_init_seconds) ** 3)

    modeldata['mass_shell'] = (10 ** modeldata['logrho']) * modeldata['volume_shell']

    merge_dfs = modeldata.merge(abundancedata, how='inner', on='inputcellid')

    print("Total mass (Msun):")
    for key in merge_dfs.keys():
        if 'X_' in key:
            merge_dfs[f'mass_{key}'] = merge_dfs[key] * merge_dfs['mass_shell'] * u.g.to('solMass')
            # get mass of element in each cell
            print(key, merge_dfs[f'mass_{key}'].sum())  # print total mass of element in solmass

    return merge_dfs


def plot_most_abundant(modelpath, args):
    model, _, _ = at.inputmodel.get_modeldata(modelpath[0], dimensions=3)
    abundances = at.inputmodel.get_initialabundances(modelpath[0])

    merge_dfs = model.merge(abundances, how='inner', on='inputcellid')
    elements = [x for x in merge_dfs.keys() if 'X_' in x]

    merge_dfs['max'] = merge_dfs[elements].idxmax(axis=1)

    merge_dfs['max'] = merge_dfs['max'].apply(lambda x: at.get_atomic_number(x[2:]))
    merge_dfs = merge_dfs[merge_dfs['max'] != 1]

    return merge_dfs


def make_3d_plot(modelpath, args):
    import pyvista as pv
    pv.set_plot_theme("document")  # set white background

    model, t_model, vmax = at.inputmodel.get_modeldata(modelpath, dimensions=3, get_abundances=False)
    abundances = at.inputmodel.get_initialabundances(modelpath)

    abundances['inputcellid'] = abundances['inputcellid'].apply(lambda x: float(x))

    merge_dfs = model.merge(abundances, how='inner', on='inputcellid')
    model = merge_dfs

    # choose what surface will be coloured by
    if args.rho:
        coloursurfaceby = 'rho'
    elif args.opacity:
        model['opacity'] = at.inputmodel.opacityinputfile.get_opacity_from_file(modelpath)
        coloursurfaceby = 'opacity'
    else:
        print(f"Colours set by X_{args.ion}")
        coloursurfaceby = f'X_{args.ion}'

    # generate grid from data
    grid = round(len(model['rho']) ** (1. / 3.))
    surfacecolorscale = np.zeros((grid, grid, grid))  # needs 3D array
    xgrid = np.zeros(grid)

    surfacearr = np.array(model[coloursurfaceby])

    i = 0
    for z in range(0, grid):
        for y in range(0, grid):
            for x in range(0, grid):
                surfacecolorscale[x, y, z] = surfacearr[i]
                xgrid[x] = -vmax + 2 * x * vmax / grid
                i += 1

    x, y, z = np.meshgrid(xgrid, xgrid, xgrid)

    mesh = pv.StructuredGrid(x, y, z)
    print(mesh)  # tells you the properties of the mesh

    mesh[coloursurfaceby] = surfacecolorscale.ravel(order='F')  # add data to the mesh
    # mesh.plot()
    minval = np.min(mesh[coloursurfaceby][np.nonzero(mesh[coloursurfaceby])])  # minimum non zero value
    print(f"{coloursurfaceby} minumin {minval}, maximum {max(mesh[coloursurfaceby])}")

    if not args.surfaces3d:
        surfacepositions = np.linspace(min(mesh[coloursurfaceby]), max(mesh[coloursurfaceby]), num=10)
        print(f"Using default surfaces {surfacepositions} \n define these with -surfaces3d for better results")
    else:
        surfacepositions = args.surfaces3d
    # surfacepositions = [1, 50, 100, 300, 500, 800, 1000, 1100, 1200, 1300, 1400, 1450, 1500] # choose these

    surf = mesh.contour(surfacepositions, scalars=coloursurfaceby)  # create isosurfaces

    surf.plot(opacity='linear', screenshot=modelpath / '3Dplot.png')    # plot surfaces and save screenshot


def addargs(parser):
    parser.add_argument('-modelpath', default=[], nargs='*', action=at.AppendPath,
                        help='Path(s) to ARTIS folder'
                        ' (may include wildcards such as * and **)')

    parser.add_argument('-o', action='store', dest='outputfile', type=Path, default=Path(),
                        help='Filename for PDF file')

    parser.add_argument('-ion', type=str, default=['Fe'], nargs='+',
                        help='Choose ion to plot. Default is Fe')

    parser.add_argument('--rho', action='store_true',
                        help='Plot rho instead of ion')

    parser.add_argument('--opacity', action='store_true',
                        help='Plot opacity from opacity.txt (if available for model)')

    parser.add_argument('-modeldim', type=int, default=None,
                        help='Choose how many dimensions. 3 for 3D, 2 for 2D')

    parser.add_argument('--plot3d', action='store_true',
                        help='Make 3D plot')

    parser.add_argument('-surfaces3d', type=float, nargs='+',
                        help='define positions of surfaces for 3D plots')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot ARTIS input model composition')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = ['.']

    args.modelpath = at.flatten_list(args.modelpath)

    if args.plot3d:
        make_3d_plot(Path(args.modelpath[0]), args)
        return

    if not args.modeldim:
        inputparams = at.get_inputparams(args.modelpath[0])
    else:
        inputparams = {'n_dimensions': args.modeldim}

    if inputparams['n_dimensions'] == 2:
        plot_2d_initial_abundances(args.modelpath, args)

    if inputparams['n_dimensions'] == 3:
        plot_3d_initial_abundances(args.modelpath, args)


if __name__ == '__main__':
    main()



import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
from pathlib import Path

import artistools as at
import artistools.estimators.estimators_classic
import artistools.inputmodel.slice1Dfromconein3dmodel
import artistools.initial_composition

from collections import namedtuple

CLIGHT = 2.99792458e10


def read_selected_mgi(modelpath, readonly_mgi, readonly_timestep=False):
    modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
    estimators = at.estimators.estimators_classic.read_classic_estimators(modelpath, modeldata,
                                                                          readonly_mgi=readonly_mgi,
                                                                          readonly_timestep=readonly_timestep)
    return estimators


def get_modelgridcells_along_axis(modelpath):

    ArgsTuple = namedtuple('args', 'modelpath sliceaxis other_axis1 other_axis2 positive_axis')
    args = ArgsTuple(
        modelpath=modelpath,
        sliceaxis='x',
        other_axis1='z',
        other_axis2='y',
        positive_axis=True
    )

    profile1d = at.inputmodel.slice1Dfromconein3dmodel.get_profile_along_axis(args=args)
    readonly_mgi = get_mgi_of_modeldata(profile1d, modelpath)

    return readonly_mgi


def get_modelgridcells_2D_slice(modeldata, modelpath):
    sliceaxis = 'z'

    slice = at.initial_composition.get_2D_slice_through_3d_model(modeldata, sliceaxis)
    readonly_mgi = get_mgi_of_modeldata(slice, modelpath)

    return readonly_mgi


def get_mgi_of_modeldata(modeldata, modelpath):
    assoc_cells, mgi_of_propcells = at.get_grid_mapping(modelpath=modelpath)
    readonly_mgi = []
    for index, row in modeldata.iterrows():
        if row['rho'] > 0:
            mgi = mgi_of_propcells[int(row['inputcellid'])-1]
            readonly_mgi.append(mgi)
    return readonly_mgi


def plot_Te_vs_time_lineofsight_3d_model(modelpath, modeldata, estimators, readonly_mgi):
    assoc_cells, mgi_of_propcells = at.get_grid_mapping(modelpath=modelpath)
    times = at.get_timestep_times_float(modelpath)

    for mgi in readonly_mgi:
        associated_modeldata_row_for_mgi = modeldata.loc[modeldata['inputcellid'] == assoc_cells[mgi][0]]

        Te = [estimators[(timestep, mgi)]['Te'] for timestep, _ in enumerate(times)]
        plt.scatter(times, Te, label=f'vel={associated_modeldata_row_for_mgi["vel_y_mid"].values[0] / CLIGHT}')

    plt.xlabel('time [days]')
    plt.ylabel('Te [K]')
    plt.xlim(0.15, 10)
    plt.xscale('log')
    plt.legend()
    plt.show()


def plot_Te_vs_velocity(modelpath, modeldata, estimators, readonly_mgi):
    assoc_cells, mgi_of_propcells = at.get_grid_mapping(modelpath=modelpath)
    times = at.get_timestep_times_float(modelpath)
    timesteps = [50, 55, 60, 65, 70, 75, 80, 90]

    for timestep in timesteps:
        Te = [estimators[(timestep, mgi)]['Te'] for mgi in readonly_mgi]

        associated_modeldata_rows = [modeldata.loc[modeldata['inputcellid'] == assoc_cells[mgi][0]] for mgi in readonly_mgi]
        velocity = [row["vel_y_mid"].values[0] / CLIGHT for row in associated_modeldata_rows]

        plt.plot(velocity, Te, label=f'{times[timestep]:.2f}', linestyle='-', marker='o')

    plt.xlabel('velocity/c')
    plt.ylabel('Te [K]')
    plt.yscale('log')
    plt.legend()
    plt.show()


def get_Te_vs_velocity_2D(modelpath, modeldata, vmax, estimators, readonly_mgi, timestep):
    assoc_cells, mgi_of_propcells = at.get_grid_mapping(modelpath=modelpath)
    times = at.get_timestep_times_float(modelpath)
    print([(ts, time) for ts, time in enumerate(times)])
    time = times[timestep]
    print(f'time {time} days')

    ngridcells = len(modeldata['inputcellid'])
    Te = np.zeros(ngridcells)

    for mgi in readonly_mgi:
        Te[assoc_cells[mgi][0]-1] = estimators[(timestep, mgi)]['Te']

    grid = round(len(modeldata['inputcellid']) ** (1./3.))
    grid_Te = np.zeros((grid, grid, grid))  # needs 3D array
    xgrid = np.zeros(grid)

    vmax = vmax / CLIGHT
    i = 0
    for z in range(0, grid):
        for y in range(0, grid):
            for x in range(0, grid):
                grid_Te[x, y, z] = Te[i]
                if modeldata['rho'][i] == 0.:
                    grid_Te[x, y, z] = None
                xgrid[x] = -vmax + 2 * x * vmax / grid
                i += 1

    return grid_Te, xgrid


def make_2d_plot(grid, grid_Te, vmax, modelpath, xgrid, time):
    # PYVISTA
    x, y, z = np.meshgrid(xgrid, xgrid, xgrid)
    mesh = pv.StructuredGrid(x, y, z)
    mesh['Te [K]'] = grid_Te.ravel(order='F')

    sargs = dict(height=0.75, vertical=True, position_x=0.02, position_y=0.1,
                 title_font_size=22, label_font_size=25)

    pv.set_plot_theme("document")  # set white background
    p = pv.Plotter()
    p.set_scale(1.5, 1.5, 1.5)
    single_slice = mesh.slice(normal='z')
    actor = p.add_mesh(single_slice, scalar_bar_args=sargs)  # , clim=[100, 60000]
    actor = p.show_bounds(grid=False, xlabel='vx / c', ylabel='vy / c', zlabel='vz / c',
                          ticks='inside', minor_ticks=False, use_2d=True, font_size=26, bold=False)

    p.camera_position = 'xy'
    p.add_title(f'{time:.1f} days')
    p.show(screenshot=modelpath / f'3Dplot_Te{time:.1f}days_disk.png')


def main():
    modelpath = Path(".")
    modeldata, _, vmax = at.inputmodel.get_modeldata(modelpath)

    # # Get mgi of grid cells along axis for 1D plot
    # # readonly_mgi = get_modelgridcells_along_axis(modelpath)
    readonly_mgi = get_modelgridcells_2D_slice(modeldata, modelpath)
    #
    timestep = 82
    times = at.get_timestep_times_float(modelpath)
    time = times[timestep]
    estimators = read_selected_mgi(modelpath, readonly_mgi=readonly_mgi, readonly_timestep=[timestep])
    grid_Te, xgrid = get_Te_vs_velocity_2D(modelpath, modeldata, vmax, estimators, readonly_mgi, timestep)
    grid = round(len(modeldata['inputcellid']) ** (1./3.))
    make_2d_plot(grid, grid_Te, vmax, modelpath, xgrid, time)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# import glob
# import itertools
import math
import os
# import sys
from pathlib import Path
from typing import Iterable

import numpy as np
import pandas as pd
from astropy import constants as const
from astropy import units as u
import matplotlib.pyplot as plt

import artistools as at
import artistools.spectra

from scipy.interpolate import interp1d


def readfile(filepath_or_buffer, modelpath=None, args=None):
    lcdata = pd.read_csv(filepath_or_buffer, delim_whitespace=True, header=None, names=['time', 'lum', 'lum_cmf'])

    if args is not None and args.gamma and modelpath is not None and at.get_inputparams(modelpath)['n_dimensions'] == 3:
        lcdata = read_3d_gammalightcurve(filepath_or_buffer)

    elif args is not None and args.plotviewingangle is not None:
        # get a list of dfs with light curves at each viewing angle
        lcdata = at.gather_res_data(lcdata, index_of_repeated_value=0)

    else:
        # the light_curve.dat file repeats x values, so keep the first half only
        lcdata = lcdata.iloc[:len(lcdata) // 2]
        lcdata.index.name = 'timestep'
    return lcdata


def read_3d_gammalightcurve(filepath_or_buffer):
    columns = ['time']
    columns.extend(np.arange(0, 100))
    lcdata = pd.read_csv(filepath_or_buffer, delim_whitespace=True, header=None)
    lcdata.columns = columns
    # lcdata = lcdata.rename(columns={0: 'time', 1: 'lum', 2: 'lum_cmf'})

    res_data = []
    for angle in np.arange(0, 100):
        res_data.append(lcdata[['time', angle]].copy())
        res_data[angle] = res_data[angle].rename(columns={angle: 'lum'})
    print(res_data)

    return res_data


def get_from_packets(modelpath, lcpath, packet_type='TYPE_ESCAPE', escape_type='TYPE_RPKT', maxpacketfiles=None):
    import artistools.packets

    packetsfiles = at.packets.get_packetsfilepaths(modelpath, maxpacketfiles=maxpacketfiles)
    nprocs_read = len(packetsfiles)
    assert nprocs_read > 0

    timearray = at.get_timestep_times_float(modelpath=modelpath, loc='mid')
    arr_timedelta = at.get_timestep_times_float(modelpath=modelpath, loc='delta')
    # timearray = np.arange(250, 350, 0.1)
    model, _, _ = at.inputmodel.get_modeldata(modelpath)
    vmax = model.iloc[-1].velocity_outer * u.km / u.s
    betafactor = math.sqrt(1 - (vmax / const.c).decompose().value ** 2)

    timearrayplusend = np.concatenate([timearray, [timearray[-1] + arr_timedelta[-1]]])

    lcdata = pd.DataFrame({'time': timearray,
                           'lum': np.zeros_like(timearray, dtype=float),
                           'lum_cmf': np.zeros_like(timearray, dtype=float)})

    for packetsfile in packetsfiles:
        dfpackets = at.packets.readfile(packetsfile, type=packet_type, escape_type=escape_type)

        if not (dfpackets.empty):
            print(f"sum of e_cmf {dfpackets['e_cmf'].sum()} e_rf {dfpackets['e_rf'].sum()}")

            binned = pd.cut(dfpackets['t_arrive_d'], timearrayplusend, labels=False, include_lowest=True)
            for binindex, e_rf_sum in dfpackets.groupby(binned)['e_rf'].sum().iteritems():
                lcdata['lum'][binindex] += e_rf_sum

            dfpackets['t_arrive_cmf_d'] = dfpackets['escape_time'] * betafactor * u.s.to('day')

            binned_cmf = pd.cut(dfpackets['t_arrive_cmf_d'], timearrayplusend, labels=False, include_lowest=True)
            for binindex, e_cmf_sum in dfpackets.groupby(binned_cmf)['e_cmf'].sum().iteritems():
                lcdata['lum_cmf'][binindex] += e_cmf_sum

    lcdata['lum'] = np.divide(lcdata['lum'] / nprocs_read * (u.erg / u.day).to('solLum'), arr_timedelta)
    lcdata['lum_cmf'] = np.divide(lcdata['lum_cmf'] / nprocs_read / betafactor * (u.erg / u.day).to('solLum'),
                                  arr_timedelta)
    return lcdata


def generate_band_lightcurve_data(modelpath, args, angle=None, modelnumber=None):
    """Method adapted from https://github.com/cinserra/S3/blob/master/src/s3/SMS.py"""
    if args and args.plotvspecpol and os.path.isfile(modelpath / 'vpkt.txt'):
        print("Found vpkt.txt, using virtual packets")
        stokes_params = at.spectra.get_specpol_data(angle, modelpath)
        vspecdata = stokes_params['I']
        timearray = vspecdata.keys()[1:]
    elif args and args.plotviewingangle and os.path.isfile(modelpath / 'specpol_res.out'):
        specfilename = os.path.join(modelpath, "specpol_res.out")
        specdataresdata = pd.read_csv(specfilename, delim_whitespace=True)
        timearray = [i for i in specdataresdata.columns.values[1:] if i[-2] != '.']
    # elif Path(modelpath, 'specpol.out').is_file():
    #     specfilename = os.path.join(modelpath, "specpol.out")
    #     specdata = pd.read_csv(specfilename, delim_whitespace=True)
    #     timearray = [i for i in specdata.columns.values[1:] if i[-2] != '.']
    else:
        specfilename = at.firstexisting(['spec.out.xz', 'spec.out.gz', 'spec.out', 'specpol.out'], path=modelpath)
        if 'specpol.out' in str(specfilename):
            specdata = pd.read_csv(specfilename, delim_whitespace=True)
            timearray = [i for i in specdata.columns.values[1:] if i[-2] != '.']  # Ignore Q and U values in pol file
        else:
            specdata = pd.read_csv(specfilename, delim_whitespace=True)
            timearray = specdata.columns.values[1:]

    filters_dict = {}
    if not args.filter:
        args.filter = ['B']

    filters_list = args.filter

    if angle is not None and os.path.isfile(modelpath / 'specpol_res.out'):
        res_specdata = at.spectra.read_specpol_res(modelpath)
        if args and args.average_every_tenth_viewing_angle:
            at.spectra.average_angle_bins(res_specdata, angle, args)

    else:
        res_specdata = None

    for filter_name in filters_list:
        if filter_name == 'bol':
            times, bol_magnitudes = bolometric_magnitude(modelpath, timearray, args, angle=angle,
                                                         res_specdata=res_specdata)
            filters_dict['bol'] = [
                (time, bol_magnitude) for time, bol_magnitude in
                zip(times, bol_magnitudes)
                if math.isfinite(bol_magnitude)]
        elif filter_name not in filters_dict:
            filters_dict[filter_name] = []

    filterdir = os.path.join(at.PYDIR, 'data/filters/')

    for filter_name in filters_list:
        if filter_name == 'bol':
            continue
        zeropointenergyflux, wavefilter, transmission, wavefilter_min, wavefilter_max \
            = get_filter_data(filterdir, filter_name)

        for timestep, time in enumerate(timearray):
            time = float(time)
            if args.timemin < time < args.timemax:
                wavelength_from_spectrum, flux = \
                    get_spectrum_in_filter_range(modelpath, timestep, time, wavefilter_min, wavefilter_max, angle,
                                                 res_specdata=res_specdata, modelnumber=modelnumber, args=args)

                if len(wavelength_from_spectrum) > len(wavefilter):
                    interpolate_fn = interp1d(wavefilter, transmission, bounds_error=False, fill_value=0.)
                    wavefilter = np.linspace(min(wavelength_from_spectrum), int(max(wavelength_from_spectrum)),
                                             len(wavelength_from_spectrum))
                    transmission = interpolate_fn(wavefilter)
                else:
                    interpolate_fn = interp1d(wavelength_from_spectrum, flux, bounds_error=False, fill_value=0.)
                    wavelength_from_spectrum = np.linspace(wavefilter_min, wavefilter_max, len(wavefilter))
                    flux = interpolate_fn(wavelength_from_spectrum)

                phot_filtobs_sn = evaluate_magnitudes(flux, transmission, wavelength_from_spectrum, zeropointenergyflux)

                # print(time, phot_filtobs_sn)
                # if phot_filtobs_sn != 0.0:
                phot_filtobs_sn = phot_filtobs_sn - 25  # Absolute magnitude
                filters_dict[filter_name].append((time, phot_filtobs_sn))

    return filters_dict


def bolometric_magnitude(modelpath, timearray, args, angle=None, res_specdata=None):
    magnitudes = []
    times = []

    if args.timemin is None or args.timemax is None:  # todo: either make it so these are define or aren't needed
        print("args.timemin or args.timemax not defined")
        quit()

    for timestep, time in enumerate(timearray):
        time = float(time)
        if args.timemin < time < args.timemax:
            if angle is not None:
                if args.plotvspecpol:
                    spectrum = at.spectra.get_vspecpol_spectrum(modelpath, time, angle, args)
                else:
                    if res_specdata is None:
                        res_specdata = at.spectra.read_specpol_res(modelpath)
                        if args and args.average_every_tenth_viewing_angle:
                            at.spectra.average_angle_bins(res_specdata, angle, args)
                    spectrum = at.spectra.get_res_spectrum(modelpath, timestep, timestep, angle=angle,
                                                           res_specdata=res_specdata)
            else:
                spectrum = at.spectra.get_spectrum(modelpath, timestep, timestep)

            integrated_flux = np.trapz(spectrum['f_lambda'], spectrum['lambda_angstroms'])
            integrated_luminosity = integrated_flux * 4 * np.pi * np.power(u.Mpc.to('cm'), 2)
            Mbol_sun = 4.74
            magnitude = Mbol_sun - (2.5 * np.log10(integrated_luminosity / const.L_sun.to('erg/s').value))
            magnitudes.append(magnitude)
            times.append(time)
            # print(const.L_sun.to('erg/s').value)
            # quit()

    return times, magnitudes


def get_filter_data(filterdir, filter_name):
    """Filter data in 'data/filters' taken from https://github.com/cinserra/S3/tree/master/src/s3/metadata"""

    with open(filterdir / Path(filter_name + '.txt'), 'r') as filter_metadata:  # defintion of the file
        line_in_filter_metadata = filter_metadata.readlines()  # list of lines

    zeropointenergyflux = float(line_in_filter_metadata[0])
    # zero point in energy flux (erg/cm^2/s)

    wavefilter, transmission = [], []
    for row in line_in_filter_metadata[4:]:
        # lines where the wave and transmission are stored
        wavefilter.append(float(row.split()[0]))
        transmission.append(float(row.split()[1]))

    wavefilter_min = min(wavefilter)
    wavefilter_max = int(max(wavefilter))

    return zeropointenergyflux, np.array(wavefilter), np.array(transmission), wavefilter_min, wavefilter_max


def get_spectrum_in_filter_range(modelpath, timestep, time, wavefilter_min, wavefilter_max, angle=None,
                                 res_specdata=None, modelnumber=None, spectrum=None, args=None):
    if spectrum is None:
        spectrum = at.spectra.get_spectrum_at_time(
            modelpath, timestep=timestep, time=time, args=args,
            angle=angle, res_specdata=res_specdata, modelnumber=modelnumber)

    wavelength_from_spectrum, flux = [], []
    for wavelength, flambda in zip(spectrum['lambda_angstroms'], spectrum['f_lambda']):
        if wavefilter_min <= wavelength <= wavefilter_max:  # to match the spectrum wavelengths to those of the filter
            wavelength_from_spectrum.append(wavelength)
            flux.append(flambda)

    return np.array(wavelength_from_spectrum), np.array(flux)


def evaluate_magnitudes(flux, transmission, wavelength_from_spectrum, zeropointenergyflux):
    cf = flux * transmission
    flux_obs = abs(np.trapz(cf, wavelength_from_spectrum))  # using trapezoidal rule to integrate
    if flux_obs == 0.0:
        phot_filtobs_sn = 0.0
    else:
        phot_filtobs_sn = -2.5 * np.log10(flux_obs / zeropointenergyflux)

    return phot_filtobs_sn


def get_band_lightcurve(band_lightcurve_data, band_name, args):

    time = [t for t, _ in band_lightcurve_data[band_name] if (args.timemin < t < args.timemax)]
    brightness_in_mag = [brightness for t, brightness in band_lightcurve_data[band_name]
                         if (args.timemin < t < args.timemax)]

    return time, brightness_in_mag


def get_colour_delta_mag(band_lightcurve_data, filter_names):
    time_dict_1 = {}
    time_dict_2 = {}

    plot_times = []
    colour_delta_mag = []

    for filter_1, filter_2 in zip(band_lightcurve_data[filter_names[0]], band_lightcurve_data[filter_names[1]]):
        # Make magnitude dictionaries where time is the key
        time_dict_1[float(filter_1[0])] = filter_1[1]
        time_dict_2[float(filter_2[0])] = filter_2[1]

    for time in time_dict_1.keys():
        if time in time_dict_2.keys():  # Test if time has a magnitude for both filters
            plot_times.append(time)
            colour_delta_mag.append(time_dict_1[time] - time_dict_2[time])

    return plot_times, colour_delta_mag


def read_hesma_lightcurve(args):
    hesma_directory = os.path.join(at.PYDIR, 'data/hesma')
    filename = args.plot_hesma_model
    hesma_modelname = hesma_directory / filename

    column_names = []
    with open(hesma_modelname) as f:
        first_line = f.readline()
        if '#' in first_line:
            for i in first_line:
                if i != '#' and i != ' ' and i != '\n':
                    column_names.append(i)

            hesma_model = pd.read_csv(hesma_modelname, delim_whitespace=True, header=None,
                                      comment='#', names=column_names)

        else:
            hesma_model = pd.read_csv(hesma_modelname, delim_whitespace=True)
    return hesma_model


def read_reflightcurve_band_data(lightcurvefilename):
    filepath = Path(at.PYDIR, 'data', 'lightcurves', lightcurvefilename)
    metadata = at.get_file_metadata(filepath)

    data_path = os.path.join(at.PYDIR, f"data/lightcurves/{lightcurvefilename}")
    lightcurve_data = pd.read_csv(data_path, comment='#')
    lightcurve_data['time'] = lightcurve_data['time'].apply(lambda x: x - (metadata['timecorrection']))
    # m - M = 5log(d) - 5  Get absolute magnitude
    if 'dist_mpc' not in metadata and 'z' in metadata:
        from astropy import cosmology
        cosmo = cosmology.FlatLambdaCDM(H0=70, Om0=0.3)
        metadata['dist_mpc'] = cosmo.luminosity_distance(metadata['z']).value
        print(f"luminosity distance from redshift = {metadata['dist_mpc']} for {metadata['label']}")

    if 'dist_mpc' in metadata:
        lightcurve_data['magnitude'] = lightcurve_data['magnitude'].apply(lambda x: (
            x - 5 * np.log10(metadata['dist_mpc'] * 10 ** 6) + 5))
    elif 'dist_modulus' in metadata:
        lightcurve_data['magnitude'] = lightcurve_data['magnitude'].apply(lambda x: (
            x - metadata['dist_modulus']))

    return lightcurve_data, metadata


def read_bol_reflightcurve_data(lightcurvefilename):
    data_path = os.path.join(at.PYDIR, f"data/lightcurves/bollightcurves/{lightcurvefilename}")
    metadata = at.get_file_metadata(data_path)

    bollightcurve_data = pd.read_csv(data_path, comment='#', delim_whitespace=True, header=None)

    bollightcurve_data = bollightcurve_data.rename(columns={0: 'time_days'})  # days
    bollightcurve_data = bollightcurve_data.rename(columns={1: 'luminosity_erg/s'})  # erg/s

    return bollightcurve_data, metadata


def get_sn_sample_bol():
    datafilepath = Path(at.PYDIR, 'data', 'lightcurves', 'SNsample', 'bololc.txt')
    sn_data = pd.read_csv(datafilepath, delim_whitespace=True, comment='#')

    print(sn_data)
    bol_luminosity = sn_data['Lmax'].astype(float)
    bol_magnitude = 4.74 - (2.5 * np.log10((10**bol_luminosity) / const.L_sun.to('erg/s').value))  # 𝑀𝑏𝑜𝑙,𝑠𝑢𝑛 = 4.74

    bol_magnitude_error_upper = bol_magnitude - (4.74 - (2.5 * np.log10((10**(bol_luminosity + sn_data['+/-.2'].astype(float))) / const.L_sun.to('erg/s').value)))
    # bol_magnitude_error_lower = (4.74 - (2.5 * np.log10((10**(bol_luminosity - sn_data['+/-.2'].astype(float))) / const.L_sun.to('erg/s').value))) - bol_magnitude
    # print(bol_magnitude_error_upper, "============")
    # print(bol_magnitude_error_lower, "============")
    # print(bol_magnitude_error_upper == bol_magnitude_error_lower)

    # a0 = plt.errorbar(x=sn_data['dm15'].astype(float), y=sn_data['Lmax'].astype(float),
    #                   yerr=sn_data['+/-.2'].astype(float), xerr=sn_data['+/-'].astype(float),
    #                   color='grey', marker='o', ls='None')
    #
    sn_data['bol_mag'] = bol_magnitude
    print(sn_data[['name', 'bol_mag', 'dm15', 'dm40']])
    sn_data[['name', 'bol_mag', 'dm15', 'dm40']].to_csv('boldata.txt', sep=' ', index=False)
    a0 = plt.errorbar(x=sn_data['dm15'].astype(float), y=bol_magnitude,
                      yerr=bol_magnitude_error_upper, xerr=sn_data['+/-'].astype(float),
                      color='k', marker='o', ls='None')

    # a0 = plt.errorbar(x=sn_data['dm15'].astype(float), y=sn_data['dm40'].astype(float),
    #                   yerr=sn_data['+/-.1'].astype(float), xerr=sn_data['+/-'].astype(float),
    #                   color='k', marker='o', ls='None')

    # a0 = plt.scatter(sn_data['dm15'].astype(float), bol_magnitude, s=80, color='k', marker='o')
    # plt.gca().invert_yaxis()
    # plt.show()

    label = 'Bolometric data (Scalzo et al. 2019)'
    return a0, label


def get_phillips_relation_data():
    datafilepath = Path(at.PYDIR, 'data', 'lightcurves', 'SNsample', 'CfA3_Phillips.dat')
    sn_data = pd.read_csv(datafilepath, delim_whitespace=True, comment='#')
    print(sn_data)

    sn_data['dm15(B)'] = sn_data['dm15(B)'].astype(float)
    sn_data['MB'] = sn_data['MB'].astype(float)

    label = 'Observed (Hicken et al. 2009)'
    return sn_data, label


def plot_phillips_relation_data():
    sn_data, label = get_phillips_relation_data()

    # a0 = plt.scatter(deltam_15B, M_B, s=80, color='grey', marker='o', label=label)
    a0 = plt.errorbar(x=sn_data['dm15(B)'], y=sn_data['MB'], yerr=sn_data['err_MB'], xerr=sn_data['err_dm15(B)'],
                      color='k', alpha=0.9, marker='.', capsize=2, label=label, ls='None', zorder=5)
    # plt.gca().invert_yaxis()
    # plt.show()
    return a0, label
